"""State module for managing Amazon Config Delivery Channel."""
import copy
from dataclasses import field
from dataclasses import make_dataclass
from typing import Any
from typing import Dict

__contracts__ = ["resource"]


async def present(
    hub,
    ctx,
    name: str,
    s3_bucket_name: str,
    resource_id: str = None,
    s3_key_prefix: str = None,
    s3_kms_key_arn: str = None,
    sns_topic_arn: str = None,
    config_snapshot_delivery_properties: make_dataclass(
        "ConfigSnapshotDeliveryProperties",
        [("delivery_frequency", str, field(default=None))],
    ) = None,
) -> Dict[str, Any]:
    """Add or update the configuration delivery channel object that delivers the configuration information to an Amazon S3 bucket and to an Amazon SNS topic.

    Args:
        name(str):
            An Idem name of the delivery channel.

        s3_bucket_name (str):
            The name of the Amazon S3 bucket to which Config delivers configuration snapshots and
            configuration history files

        resource_id (str, Optional):
            The resource Id of the delivery channel.

        s3_key_prefix (str, Optional):
            The prefix for the specified Amazon S3 bucket.

        s3_kms_key_arn (str, Optional):
            The Amazon Resource Name (ARN) of the Key Management Service (KMS ) KMS key (KMS key) used to
            encrypt objects delivered by Config.

        sns_topic_arn (str, Optional):
            The Amazon Resource Name (ARN) of the Amazon SNS topic to which Config sends notifications
            about configuration changes.

        config_snapshot_delivery_properties (dict[str, Any], Optional):
            The options for how often Config delivers configuration snapshots to the Amazon S3 bucket.

            * delivery_frequency (str, Optional):
              The frequency with which Config delivers configuration snapshots.

    Request syntax:
        .. code-block:: sls

            [aws-config-delivery_channel]:
              aws.config.delivery_channel.present:
              - name: 'string'
              - resource_id: 'string'
              - s3_bucket_name: 'string'
              - s3_key_prefix: 'string'
              - s3_kms_key_arn: 'string'
              - sns_topic_arn: 'string'
              - config_snapshot_delivery_properties: dict
                delivery_frequency: 'string'

    Returns:
         Dict[str, Any]

    Examples:
        .. code-block:: sls

            aws-config-delivery_channel:
              aws.config.delivery_channel.present:
                - name: 'delivery-channel'
                - resource_id: 'delivery-channel'
                - s3_bucket_name: 'Test-S3'
                - s3_key_prefix: 'S3-prefix'
                - s3_kms_key_arn: 'Test-Kms-Key'
                - sns_topic_arn: 'Test-Topic'
                - config_snapshot_delivery_properties:
                  delivery_frequency: 'One_Hour'
    """
    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)
    before = None
    resource_updated = False
    update_payload = {}
    update_config_snapshot_delivery_properties = None
    resource_id = resource_id if resource_id else name
    before = await hub.exec.boto3.client.config.describe_delivery_channels(
        ctx, DeliveryChannelNames=[resource_id]
    )
    if before:
        result[
            "old_state"
        ] = await hub.tool.aws.config.conversion_utils.convert_raw_config_delivery_channel_to_present(
            ctx,
            raw_resource=before["ret"]["DeliveryChannels"][0],
            idem_resource_name=name,
        )
        plan_state = copy.deepcopy(result["old_state"])
        # updating config_snapshot_delivery_properties
        if config_snapshot_delivery_properties is not None:
            update_config_snapshot_delivery_properties = (
                config_snapshot_delivery_properties
            )
        update_payload = await hub.tool.aws.config.delivery_channel.get_updated_payload_delivery_channel(
            ctx,
            resource_id=resource_id,
            before=before["ret"]["DeliveryChannels"][0],
            s3_bucket_name=s3_bucket_name,
            s3_key_prefix=s3_key_prefix,
            s3_kms_key_arn=s3_kms_key_arn,
            sns_topic_arn=sns_topic_arn,
            config_snapshot_delivery_properties=update_config_snapshot_delivery_properties,
        )

        if update_payload["ret"]:
            resource_updated = True
            if not ctx.get("test", False):
                update_ret = await hub.exec.boto3.client.config.put_delivery_channel(
                    ctx=ctx, DeliveryChannel=update_payload["ret"]
                )
                if not update_ret["result"]:
                    result["comment"] = result["comment"] + update_ret["comment"]
                    result["result"] = False
                    return result

                result["comment"] = (f"Updated aws.config.delivery_channel '{name}'",)
                result["result"] = update_ret["result"]
            else:
                for key, value in update_payload["ret"].items():
                    plan_state[key] = value

        if not result["result"]:
            result["comment"] = result["comment"] + (f"{name} already exists",)
    else:
        delivery_props = {}
        dc_payload = {"name": name, "s3BucketName": s3_bucket_name}
        if sns_topic_arn:
            dc_payload["snsTopicARN"] = sns_topic_arn
        if s3_key_prefix:
            dc_payload["s3KeyPrefix"] = s3_key_prefix
        if s3_kms_key_arn:
            dc_payload["s3KmsKeyArn"] = s3_kms_key_arn
        if config_snapshot_delivery_properties:
            delivery_props["deliveryFrequency"] = config_snapshot_delivery_properties[
                "delivery_frequency"
            ]
            dc_payload["configSnapshotDeliveryProperties"] = delivery_props
        if ctx.get("test", False):
            result["new_state"] = hub.tool.aws.test_state_utils.generate_test_state(
                enforced_state={}, desired_state=dc_payload
            )
            result["comment"] = (f"Would create aws.config.delivery_channel '{name}'",)
            return result

        ret = await hub.exec.boto3.client.config.put_delivery_channel(
            ctx, DeliveryChannel=dc_payload
        )
        result["result"] = ret["result"]
        if not result["result"]:
            result["comment"] = ret["comment"] + ret["comment"]
            return result
            # resource_id = name
        result["comment"] = result["comment"] + (
            f"Created aws.config.delivery_channel '{name}'",
        )

    if ctx.get("test", False):
        result["new_state"] = plan_state
        if not resource_updated:
            result["comment"] = f"Would create aws.config.delivery_channel '{name}'"
        else:
            result["comment"] = (f"Would update aws.config.delivery_channel '{name}'",)
    elif (not before) or resource_updated:
        after = await hub.exec.boto3.client.config.describe_delivery_channels(
            ctx, DeliveryChannelNames=[resource_id]
        )
        result[
            "new_state"
        ] = await hub.tool.aws.config.conversion_utils.convert_raw_config_delivery_channel_to_present(
            ctx,
            raw_resource=after["ret"]["DeliveryChannels"][0],
            idem_resource_name=name,
        )
    else:
        result["new_state"] = copy.deepcopy(result["old_state"])
    return result


async def absent(
    hub,
    ctx,
    name: str,
    resource_id: str = None,
) -> Dict[str, Any]:
    """Deletes the specified delivery channel.

    Before you can delete the delivery channel, you must stop the configuration recorder by using the StopConfigurationRecorder action.

    Args:
        name(str): An Idem name of the rule.
        resource_id(str, Optional): AWS Config delivery channel Name. Idem automatically considers this resource being absent
         if this field is not specified.

    Returns:
          Dict[str, Any]

    Examples:
          .. code-block:: sls

            ec2-instance-no-public-ip:
              aws.config.delivery_channel.absent:
              - name: ec2-instance-no-public-ip
              - resource_id: ec2-instance-no-public-ip

    """
    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)
    resource_id = resource_id if resource_id else name
    # if not resource_id:
    #     result["comment"] = hub.tool.aws.comment_utils.already_absent_comment(
    #         resource_type="aws.config.delivery_channel", name=name
    #     )
    #     return result
    before = await hub.exec.boto3.client.config.describe_delivery_channels(
        ctx, DeliveryChannelNames=[resource_id]
    )
    if not before:
        result["comment"] = hub.tool.aws.comment_utils.already_absent_comment(
            resource_type="aws.config.delivery_channel", name=name
        )
    elif ctx.get("test", False):
        result[
            "old_state"
        ] = await hub.tool.aws.config.conversion_utils.convert_raw_config_delivery_channel_to_present(
            ctx,
            raw_resource=before["ret"]["DeliveryChannels"][0],
            idem_resource_name=resource_id,
        )
        result["comment"] = hub.tool.aws.comment_utils.would_delete_comment(
            resource_type="aws.config.delivery_channel", name=resource_id
        )
        return result
    else:
        try:
            result[
                "old_state"
            ] = await hub.tool.aws.config.conversion_utils.convert_raw_config_delivery_channel_to_present(
                ctx,
                raw_resource=before["ret"]["DeliveryChannels"][0],
                idem_resource_name=resource_id,
            )

            ret = await hub.exec.boto3.client.config.delete_delivery_channel(
                ctx, DeliveryChannelName=resource_id
            )
            result["result"] = ret["result"]
            if not result["result"]:
                result["comment"] = ret["comment"]
                return result
            result["comment"] = hub.tool.aws.comment_utils.delete_comment(
                resource_type="aws.config.delivery_channel", name=name
            )
        except hub.tool.boto3.exception.ClientError as e:
            result["comment"] = result["comment"] + (f"{e.__class__.__name__}: {e}",)
            result["result"] = False
    return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    """Describe the resource in a way that can be recreated/managed with the corresponding "present" function.

    Return details about your delivery channel.

    Returns:
        Dict[str, Any]

    Examples:
        .. code-block:: bash

            $ idem describe aws.config.delivery_channel
    """
    result = {}
    ret = await hub.exec.boto3.client.config.describe_delivery_channels(ctx)
    if not ret["result"]:
        hub.log.debug(f"Could not describe Delivery Channels {ret['comment']}")
        return {}

    for resource in ret["ret"]["DeliveryChannels"]:
        resource_id = resource.get("name")
        resource_translated = await hub.tool.aws.config.conversion_utils.convert_raw_config_delivery_channel_to_present(
            ctx, raw_resource=resource, idem_resource_name=resource_id
        )
        result[resource_id] = {
            "aws.config.delivery_channel.present": [
                {parameter_key: parameter_value}
                for parameter_key, parameter_value in resource_translated.items()
            ]
        }
    return result



import copy
from collections import OrderedDict
from typing import Any
from typing import Dict


def convert_raw_function_event_invoke_config_to_present(
    hub,
    raw_resource: Dict[str, Any],
    function_name: str = None,
    idem_resource_name: str = None,
) -> Dict[str, Any]:
    r"""
    Convert AWS returned data structure to correct idem function_event_invoke_config present state

     Args:
        hub: required for functions in hub
        raw_resource(Dict): The aws response to convert
        function_name (string, Optional): The name of the Lambda function
        idem_resource_name(string, optional): An Idem name of the resource.

    Returns: Valid idem state for function_event_invoke_config of type Dict['string', Any]
    """
    function_arn = raw_resource.get("FunctionArn")
    resource_parameters = OrderedDict(
        {
            "MaximumRetryAttempts": "maximum_retry_attempts",
            "MaximumEventAgeInSeconds": "maximum_event_age_in_seconds",
            "DestinationConfig": "destination_config",
        }
    )
    resource_translated = {
        "name": idem_resource_name if idem_resource_name else function_arn,
        "resource_id": function_arn,
        "function_name": function_name,
    }
    for parameter_raw, parameter_present in resource_parameters.items():
        if raw_resource.get(parameter_raw) is not None:
            resource_translated[parameter_present] = raw_resource.get(parameter_raw)

    qualifier = hub.tool.aws.arn_utils.get_qualifier(function_arn)
    if qualifier:
        resource_translated["qualifier"] = qualifier

    return resource_translated


async def update_function_event_invoke_config(
    hub,
    ctx,
    function_name: str,
    current_state: Dict[str, Any],
    desired_state: Dict[str, Any],
):
    r"""
    Updates configuration options for asynchronous invocation on a function

    Args:
        hub: required for functions in hub.
        ctx: context.
        function_name(string): The name of the Lambda function
        current_state(Dict): Previous state of the rest_api resource
        desired_state(Dict): Parameters from SLS file.

    Returns:
        Dict[str, Any]
    """
    modify_params = dict(
        {
            "Qualifier": "qualifier",
            "MaximumRetryAttempts": "maximum_retry_attempts",
            "MaximumEventAgeInSeconds": "maximum_event_age_in_seconds",
            "DestinationConfig": "destination_config",
        }
    )
    result = dict(comment=(), result=True, ret=None)
    plan_state = copy.deepcopy(current_state)
    params_to_modify = {}

    # create a dict 'params_to_modify' of raw parameter key mapped to desired value,
    # where the desired value is not none and current value does not match desired value
    for param_raw, param_present in modify_params.items():
        if desired_state.get(param_present) is not None and current_state.get(
            param_present
        ) != desired_state.get(param_present):
            params_to_modify[param_raw] = desired_state[param_present]

    if params_to_modify:
        for key, value in params_to_modify.items():
            plan_state[modify_params.get(key)] = value
        if ctx.get("test", False):
            result["comment"] = hub.tool.aws.comment_utils.would_update_comment(
                resource_type="aws.lambda.function_event_invoke_config",
                name=current_state["name"],
            )
        else:
            update_ret = await hub.exec.boto3.client[
                "lambda"
            ].update_function_event_invoke_config(
                ctx, FunctionName=function_name, **params_to_modify
            )
            if not update_ret["result"]:
                result["result"] = False
                result["comment"] = update_ret["comment"]
                return result
            result["comment"] = hub.tool.aws.comment_utils.update_comment(
                resource_type="aws.lambda.function_event_invoke_config",
                name=current_state["name"],
            )
        result["ret"] = plan_state
    else:
        result["comment"] = hub.tool.aws.comment_utils.already_exists_comment(
            resource_type="aws.lambda.function_event_invoke_config",
            name=current_state["name"],
        )

    return result



import copy
from collections import OrderedDict
from typing import Any
from typing import Dict
from typing import List


async def convert_raw_integration_to_present(
    hub,
    raw_resource: Dict[str, Any],
    resource_id: str,
    idem_resource_name: str = None,
) -> Dict[str, Any]:

    resource_parameters = OrderedDict(
        {
            "type": "input_type",
            "passthroughBehavior": "passthrough_behavior",
            "timeoutInMillis": "timeout_in_millis",
            "cacheNamespace": "cache_namespace",
            "cacheKeyParameters": "cache_key_parameters",
        }
    )

    rest_api_id, parent_resource_id, http_method = resource_id.split("-")

    resource_translated = {
        "resource_id": resource_id,
        "rest_api_id": rest_api_id,
        "parent_resource_id": parent_resource_id,
        "http_method": http_method,
    }

    if idem_resource_name:
        resource_translated["name"] = idem_resource_name

    for parameter_raw, parameter_present in resource_parameters.items():
        if raw_resource.get(parameter_raw):
            resource_translated[parameter_present] = raw_resource.get(parameter_raw)

    return resource_translated


async def update(
    hub,
    ctx,
    old_state: Dict[str, Any],
    updateable_parameters: Dict[str, Any],
):
    r"""

    Updates an API Gateway Integration.

    Args:
        hub:
        ctx:
        old_state(Dict): Previous state of the resource.
        updateable_parameters(Dict): Parameters from SLS File.

    Returns:
        Dict[str, Any]

    """

    result = dict(comment=(), result=True, ret=None)
    patch_ops = []

    new_state = copy.deepcopy(old_state)
    updated = False

    for key, value in updateable_parameters.items():
        if key == "passthrough_behavior":
            key_final = "passthroughBehavior"
        elif key == "connection_id":
            key_final = "connectionId"
        elif key == "cache_key_parameters":
            key_final = "cacheKeyParameters"
        elif key == "cache_namespace":
            key_final = "cacheNamespace"
        elif key == "timeout_in_millis":
            key_final = "timeoutInMillis"
        elif key == "input_type":
            key_final = "type"
        if old_state.get(key) and value is not None and value != old_state.get(key):
            patch_ops.append(
                {
                    "op": "replace",
                    "path": f"/{key_final}",
                    "value": str(value)
                    if not isinstance(value, List)
                    else ",".join(value),
                }
            )
            updated = True
        elif old_state.get(key) is None and value is not None:
            patch_ops.append(
                {
                    "op": "add",
                    "path": f"/{key_final}",
                    "value": str(value)
                    if not isinstance(value, List)
                    else ",".join(value),
                }
            )
            new_state[key] = value
            updated = True

    if ctx.get("test", False):
        if updated:
            result["comment"] = hub.tool.aws.comment_utils.would_update_comment(
                resource_type="aws.apigateway.integration", name=old_state["name"]
            )
            result["ret"] = new_state
        else:
            result["comment"] = hub.tool.aws.comment_utils.already_exists_comment(
                resource_type="aws.apigateway.integration", name=old_state["name"]
            )
        return result

    if updated:
        update_ret = await hub.exec.boto3.client.apigateway.update_integration(
            ctx,
            restApiId=old_state["rest_api_id"],
            resourceId=old_state["parent_resource_id"],
            httpMethod=old_state["http_method"],
            patchOperations=patch_ops,
        )

        if not update_ret["result"]:
            result["result"] = False
            result["comment"] = update_ret["comment"]
            return result
        else:
            result["comment"] = hub.tool.aws.comment_utils.update_comment(
                resource_type="aws.apigateway.integration", name=old_state["name"]
            )
            result["ret"] = new_state
    else:
        result["comment"] = hub.tool.aws.comment_utils.already_exists_comment(
            resource_type="aws.apigateway.integration", name=old_state["name"]
        )

    return result



import copy
from collections import OrderedDict
from typing import Any
from typing import Dict
from typing import List

STATE_NAME = "aws.neptune.db_instance"


async def search_raw(
    hub, ctx, name: str, resource_id: str = None, filters: List = None
) -> Dict:
    """
    Fetch one or more neptune db instances from AWS. The return will be in the same format as what the boto3 api returns.

    Args:
        name(string): The name of the Idem state.
        resource_id(string, optional): AWS Neptune DBInstanceIdentifier to identify the resource.
        filters(list, optional): A filter that specifies one or more DB instances to describe.
            Supported filters:
            db-cluster-id - Accepts DB cluster identifiers and DB cluster Amazon Resource Names (ARNs). The results list will only include information about the DB instances associated with the DB clusters identified by these ARNs.
            engine - Accepts an engine name (such as neptune ), and restricts the results list to DB instances created by that engine.

    Returns:
        {"result": True|False, "comment": A message List, "ret": Dict}
    """
    result = dict(comment=[], ret=None, result=True)
    syntax_validation = hub.tool.aws.search_utils.search_filter_syntax_validation(
        filters=filters
    )
    if not syntax_validation["result"]:
        result["result"] = False
        result["comment"] = list(syntax_validation["comment"])
        return result
    boto3_filter = hub.tool.aws.search_utils.convert_search_filter_to_boto3(
        filters=filters
    )
    ret = await hub.exec.boto3.client.neptune.describe_db_instances(
        ctx,
        DBInstanceIdentifier=resource_id if resource_id else None,
        Filters=boto3_filter,
    )
    result["result"] = ret["result"]
    result["comment"] = list(ret["comment"])
    result["ret"] = ret["ret"]
    return result


async def update(
    hub,
    ctx,
    name: str,
    resource_id: str,
    desired_state: Dict,
    is_created: bool,
    timeout: Dict,
) -> Dict[str, Any]:
    """
    Updates the neptune db instance represented by the provided resource_id to the desired state.
    If is_created flag is set to True, the comment in result is not updated as it is assumed that
    the resource is being updated right after it was created to apply remaining parameters.
    """
    # below is the mapping of boto3 string parameters applicable for modification of neptune db_instance in idem
    # it contains all accepted string parameters
    modify_str_params = OrderedDict(
        {
            "DBInstanceClass": "db_instance_class",
            "DBSubnetGroupName": "db_subnet_group_name",
            "MasterUserPassword": "master_user_password",
            "DBParameterGroupName": "db_parameter_group_name",
            "BackupRetentionPeriod": "backup_retention_period",
            "PreferredBackupWindow": "preferred_backup_window",
            "PreferredMaintenanceWindow": "preferred_maintenance_window",
            "EngineVersion": "engine_version",
            "LicenseModel": "license_model",
            "OptionGroupName": "option_group_name",
            "StorageType": "storage_type",
            "TdeCredentialArn": "tde_credential_arn",
            "TdeCredentialPassword": "tde_credential_password",
            "CACertificateIdentifier": "ca_certificate_identifier",
            "Domain": "domain",
            "MonitoringRoleArn": "monitoring_role_arn",
            "DomainIAMRoleName": "domain_iam_role_name",
            "PerformanceInsightsKMSKeyId": "performance_insights_kms_key_id",
        }
    )
    # below is mapping of boto3 list params applicable for modifying neptune db_instance
    modify_list_of_str_params = {
        "DBSecurityGroups": "db_security_groups",
        "VpcSecurityGroupIds": "vpc_security_group_ids",
    }
    # below is mapping of boto3 int params applicable for modifying neptune db_instance
    modify_int_params = {
        "AllocatedStorage": "allocated_storage",
        "DBPortNumber": "port",
        "Iops": "iops",
        "MonitoringInterval": "monitoring_interval",
        "PromotionTier": "promotion_tier",
    }
    # below is mapping of boto3 bool params applicable for modifying neptune db_instance
    modify_bool_params = {
        "MultiAZ": "multi_az",
        "ApplyImmediately": "apply_immediately",
        "AllowMajorVersionUpgrade": "allow_major_version_upgrade",
        "AutoMinorVersionUpgrade": "auto_minor_version_upgrade",
        "PubliclyAccessible": "publicly_accessible",
        "CopyTagsToSnapshot": "copy_tags_to_snapshot",
        "EnableIAMDatabaseAuthentication": "enable_iam_database_authentication",
        "EnablePerformanceInsights": "enable_performance_insights",
        "DeletionProtection": "deletion_protection",
    }

    result = dict(comment=(), old_state=None, new_state=None, result=True)
    if ctx.get("test", False) and is_created:
        current_state = desired_state
    else:
        current_state_result = await hub.exec.aws.neptune.db_instance.get(
            ctx=ctx, name=name, resource_id=resource_id
        )
        if not (
            current_state_result
            and current_state_result["result"]
            and current_state_result["ret"]
        ):
            result["result"] = False
            result["comment"] = current_state_result["comment"]
            return result
        current_state = current_state_result["ret"]
    if not is_created:
        result["old_state"] = current_state
    resource_arn = current_state.get("db_instance_arn")
    params_to_modify = {}
    plan_state = copy.deepcopy(current_state)

    # create a dict 'params_to_modify' of raw parameter key of type string mapped to desired string value,
    # where the desired value is not none and current value does not match desired value

    # process all string parameters
    for param_raw, param_present in modify_str_params.items():
        if type(desired_state.get(param_present)) == str and current_state.get(
            param_present
        ) != desired_state.get(param_present):
            params_to_modify[param_raw] = desired_state[param_present]
            plan_state[param_present] = desired_state[param_present]

    # process list[string] data
    for param_raw, param_present in modify_list_of_str_params.items():
        if type(desired_state.get(param_present)) == list and set(
            current_state.get(param_present)
        ) != set(desired_state.get(param_present)):
            params_to_modify[param_raw] = desired_state[param_present]
            plan_state[param_present] = desired_state[param_present]

    # process integer params
    for param_raw, param_present in modify_int_params.items():
        if type(desired_state.get(param_present)) == int and current_state.get(
            param_present
        ) != desired_state.get(param_present):
            params_to_modify[param_raw] = desired_state[param_present]
            plan_state[param_present] = desired_state[param_present]

    # process bools
    for param_raw, param_present in modify_bool_params.items():
        if type(desired_state.get(param_present)) == bool and current_state.get(
            param_present
        ) != desired_state.get(param_present):
            params_to_modify[param_raw] = desired_state[param_present]
            plan_state[param_present] = desired_state[param_present]

    # process CloudwatchLogsExportConfiguration as a special case, since this parameter is not part of describe and is not returned by AWS
    # If user specifies this, just pass it along to boto3 and consider this resource to be updated
    if desired_state.get("CloudwatchLogsExportConfiguration"):
        params_to_modify["CloudwatchLogsExportConfiguration"] = desired_state[
            "cloudwatch_logs_export_configuration"
        ]
        plan_state[param_present] = desired_state[param_present]

    # if we have some parameters to modify, apply it
    if params_to_modify:
        if ctx.get("test", False):
            if not is_created:
                result["comment"] = hub.tool.aws.comment_utils.would_update_comment(
                    resource_type=STATE_NAME, name=name
                )
        else:
            modify_db_instance_result = (
                await hub.exec.boto3.client.neptune.modify_db_instance(
                    ctx, DBInstanceIdentifier=resource_id, **params_to_modify
                )
            )
            if not modify_db_instance_result["result"]:
                result["comment"] = modify_db_instance_result["comment"]
                result["result"] = False
                return result
            # wait for update of instance
            waiter_config = hub.tool.aws.waiter_utils.create_waiter_config(
                default_delay=15,
                default_max_attempts=40,
                timeout_config=timeout.get("update") if timeout else None,
            )
            try:
                await hub.tool.boto3.client.wait(
                    ctx,
                    "neptune",
                    "db_instance_available",
                    None,
                    30,
                    DBInstanceIdentifier=resource_id,
                    WaiterConfig=waiter_config,
                )
            except Exception as e:
                result["comment"] = result["comment"] + (str(e),)
                result["result"] = False
            if not is_created:
                result["comment"] = result[
                    "comment"
                ] + hub.tool.aws.comment_utils.update_comment(
                    resource_type=STATE_NAME, name=name
                )

    # modify tags if it has changed
    old_tags = current_state.get("tags")
    new_tags = desired_state.get("tags")
    if new_tags is not None and new_tags != old_tags:
        # below code updates tags on AWS only if it is not a test run
        update_tags_ret = await hub.tool.aws.neptune.tag.update_tags(
            ctx=ctx,
            resource_arn=resource_arn,
            old_tags=old_tags,
            new_tags=new_tags,
        )
        if not update_tags_ret["result"]:
            result["comment"] = update_tags_ret["comment"]
            result["result"] = False
            hub.log.debug(f"Failed updating tags for aws.neptune.db_instance '{name}'")
            return result
        # update comment if tags were updated
        if not is_created:
            result["comment"] = hub.tool.aws.comment_utils.update_comment(
                resource_type=STATE_NAME, name=name
            )
            result["comment"] += update_tags_ret["comment"]

        if ctx.get("test", False) and update_tags_ret["ret"] is not None:
            plan_state["tags"] = update_tags_ret["ret"]
            if not is_created:
                result["comment"] = hub.tool.aws.comment_utils.would_update_comment(
                    resource_type=STATE_NAME, name=resource_id
                )
                result["comment"] += update_tags_ret["comment"]

    # set new_state
    if ctx.get("test", False):
        result["new_state"] = plan_state
    else:
        current_db_instance_result = await hub.exec.aws.neptune.db_instance.get(
            ctx=ctx, name=name, resource_id=resource_id
        )
        if not (
            current_db_instance_result
            and current_db_instance_result["result"]
            and current_db_instance_result["ret"]
        ):
            result["result"] = False
            result["comment"] += current_db_instance_result["comment"]
            return result
        result["new_state"] = current_db_instance_result["ret"]

    return result



"""Exec module for managing EC2 FlowLogs."""
from typing import Dict
from typing import List

__func_alias__ = {"list_": "list"}


async def get(
    hub,
    ctx,
    name,
    resource_id: str = None,
    filters: List = None,
) -> Dict:
    """Get a flow_log resource from AWS.

    If more than one resource is found, the first resource returned from AWS
    will be used. The function returns None when no resource is found.

    Args:
        name(str):
            The name of the Idem state.

        resource_id(str, optional):
            Aws flow_log id to identify the resource.

        filters(list, optional):
            One or more filters: for example, tag :<key>, tag-key. A complete list of filters can be found at
            https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/ec2.html#EC2.Client.describe_flow_logs
    """
    result = dict(comment=[], ret=None, result=True)
    ret = await hub.tool.aws.ec2.flow_log.search_raw(
        ctx=ctx,
        name=name,
        resource_id=resource_id,
        filters=filters,
    )
    if not ret["result"]:
        if "InvalidParameterValue" in str(ret["comment"]):
            result["comment"].append(
                hub.tool.aws.comment_utils.get_empty_comment(
                    resource_type="aws.ec2.flow_log", name=name
                )
            )
            result["comment"] += list(ret["comment"])
            return result
        result["comment"] += list(ret["comment"])
        result["result"] = False
        return result
    if not ret["ret"]["FlowLogs"]:
        result["comment"].append(
            hub.tool.aws.comment_utils.get_empty_comment(
                resource_type="aws.ec2.flow_log", name=name
            )
        )
        return result

    resource = ret["ret"]["FlowLogs"][0]
    if len(ret["ret"]["FlowLogs"]) > 1:
        result["comment"].append(
            f"More than one aws.ec2.flow_log resource was found. Use resource {resource.get('FlowLogId')}"
        )
    result["ret"] = hub.tool.aws.ec2.conversion_utils.convert_raw_flow_log_to_present(
        resource
    )
    return result


async def list_(hub, ctx, name, filters: List = None) -> Dict:
    """Fetch a list of flow_logs from AWS.

    The function returns empty list when no resource is found.

    Args:
        name(str):
            The name of the Idem state.

        filters(list, optional):
            One or more filters: for example, tag :<key>, tag-key. A complete list of filters can be found at
            https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/ec2.html#EC2.Client.describe_flow_logs
    """
    result = dict(comment=[], ret=[], result=True)
    ret = await hub.tool.aws.ec2.flow_log.search_raw(
        ctx=ctx,
        name=name,
        filters=filters,
    )
    if not ret["result"]:
        result["comment"] += list(ret["comment"])
        result["result"] = False
        return result
    if not ret["ret"]["FlowLogs"]:
        result["comment"].append(
            hub.tool.aws.comment_utils.list_empty_comment(
                resource_type="aws.ec2.flow_log", name=name
            )
        )
        return result
    for flow_log in ret["ret"]["FlowLogs"]:
        result["ret"].append(
            hub.tool.aws.ec2.conversion_utils.convert_raw_flow_log_to_present(flow_log)
        )
    return result



from typing import List

# Monitoring isn't supported by localstack
__update_on_fail__ = "continue"


async def apply(
    hub, ctx, resource, *, old_value: bool, new_value: bool, comments: List[str]
) -> bool:
    """
    Modify an ec2 instance based on a single parameter in it's "present" state

    Args:
        hub:
        ctx: The ctx from a state module call
        resource: An ec2 instance resource object
        old_value: The previous value from the attributes of an existing instance
        new_value: The desired value from the ec2 instance present state parameters
        comments: A running list of comments abound the update process
    """
    if new_value is True:
        # Start monitoring the instance
        ret = await hub.exec.boto3.client.ec2.monitor_instances(
            ctx, InstanceIds=[resource.id]
        )
        if ret.comment:
            comments.append(ret.comment)
        return ret.result
    elif new_value is False:
        # Stop monitoring the instance
        ret = await hub.exec.boto3.client.ec2.unmonitor_instances(
            ctx, InstanceIds=[resource.id]
        )
        await hub.tool.boto3.resource.exec(resource, "wait_until_stopped")

        if ret.comment:
            comments.append(ret.comment)
        return ret.result



from dataclasses import field
from dataclasses import make_dataclass
from typing import Any
from typing import Dict
from typing import List


async def get_all_hosted_zones(hub, ctx) -> Dict[str, Any]:
    """
    Describes all the hosted zones available.

    Args:
        hub:
        ctx:

    Returns:
        {"result": True|False, "comment": "A message Tuple", "ret": None}
    """
    result = dict(comment=(), result=True, ret=None)
    hosted_zones = []
    ret = await hub.exec.boto3.client.route53.list_hosted_zones(ctx)
    if not ret["result"]:
        result["result"] = ret["result"]
        result["comment"] = ret["comment"]
        return result
    for hosted_zone in ret["ret"]["HostedZones"]:
        resource_id = hosted_zone.get("Id")
        ret = await hub.exec.boto3.client.route53.get_hosted_zone(ctx, Id=resource_id)
        if not ret["result"]:
            result["result"] = ret["result"]
            result["comment"] = ret["comment"]
            return result

        temp_hosted_zone = ret["ret"]["HostedZone"]

        if ret["ret"].get("VPCs"):
            temp_hosted_zone["VPCs"] = ret["ret"].get("VPCs")
        if ret["ret"].get("DelegationSet"):
            temp_hosted_zone["DelegationSet"] = ret["ret"].get("DelegationSet")

        tags_ret = await hub.exec.boto3.client.route53.list_tags_for_resource(
            ctx, ResourceType="hostedzone", ResourceId=resource_id.split("/")[-1]
        )
        if not tags_ret["result"]:
            result["result"] = tags_ret["result"]
            result["comment"] = tags_ret["comment"]
            return result

        if tags_ret["result"] and tags_ret["ret"].get("ResourceTagSet").get("Tags"):
            temp_hosted_zone["Tags"] = tags_ret["ret"]["ResourceTagSet"].get("Tags")
        hosted_zones.append(temp_hosted_zone)

    result["ret"] = {"HostedZones": hosted_zones}
    return result


def get_hosted_zones_with_filters(
    hub,
    raw_hosted_zones: List,
    hosted_zone_name: str = None,
    private_zone: bool = None,
    vpc_id: str = None,
    tags: Dict[str, Any]
    or List[
        make_dataclass("Tag", [("Key", str), ("Value", str, field(default=None))])
    ] = None,
):
    """
    Returns the hosted_zone with the specified filters, if it is available.

    Args:
        raw_hosted_zones(Dict): Dict of all the described hosted_zones
        hosted_zone_name(string, optional): Domain name of hosted_zone to filter.
        private_zone(bool, optional): Bool argument to specify a private hosted_zone. One of the filter option for hosted_zone
        vpc_id(string, optional): The vpc_id associated with the hosted_zone. One of the filter option for hosted_zone
        tags(List, optional): Tags of the hosted_zone. One of the filter option for hosted_zone

    """
    result = dict(comment=(), result=True, ret=None)
    hosted_zones = raw_hosted_zones["HostedZones"]

    # filter_hosted_zones() returns True if all the filters match for a hosted_zone, and it is added to the list
    filtered_hosted_zones = list(
        filter(
            lambda x: filter_hosted_zones(
                x, hosted_zone_name, private_zone, vpc_id, tags
            ),
            hosted_zones,
        )
    )

    if not filtered_hosted_zones:
        result["comment"] = (
            f"Unable to find aws.route53.hosted_zone resource with given filters",
        )
        return result

    result["comment"] = (
        f"Found this {len(filtered_hosted_zones)} aws.route53.hosted_zone resource(s) with given filters",
    )

    result["ret"] = []
    for filtered_hosted_zone in filtered_hosted_zones:
        # Building the format in which hosted_zone conversion_utils takes the hosted_zone resource
        result["ret"].append(
            {
                "ret": {
                    "HostedZone": filtered_hosted_zone,
                    "VPCs": filtered_hosted_zone.get("VPCs"),
                    "DelegationSet": filtered_hosted_zone.get("DelegationSet"),
                },
                "tags": filtered_hosted_zone.get("Tags"),
            }
        )

    return result


def filter_hosted_zones(
    hosted_zone: Dict,
    hosted_zone_name: str = None,
    private_zone: bool = None,
    vpc_id: str = None,
    tags: List = None,
):
    """
    Returns True if the hosted_zone checks all the filters provided or return False

    Args:
        hosted_zone(Dict): The described hosted_zone
        hosted_zone_name(string, optional): Domain name of hosted_zone to filter.
        private_zone(bool, optional): Bool argument to specify a private hosted_zone. One of the filter option for hosted_zone
        vpc_id(string, optional): The vpc_id associated with the hosted_zone. One of the filter option for hosted_zone
        tags(List, optional): Tags of the hosted_zone. One of the filter option for hosted_zone

    """
    # Return True if all the provided filters match or return False.

    if hosted_zone_name:
        if hosted_zone["Name"] != hosted_zone_name:
            return False

    if private_zone is not None:
        if hosted_zone["Config"]["PrivateZone"] != private_zone:
            return False

    if vpc_id:
        found = False
        if hosted_zone["VPCs"]:
            for vpc in hosted_zone["VPCs"]:
                if vpc["VPCId"] == vpc_id:
                    found = True
                    break
            if not found:
                return False

    # Checking if all the tags in the filter match with the tags present in the hosted_zone.If not we return False
    if tags:
        tags2 = hosted_zone.get("Tags")
        if tags2 is None:
            return False
        tags2_map = {tag.get("Key"): tag for tag in tags2}
        for tag in tags:
            if tag["Key"] not in tags2_map or (
                tags2_map.get(tag["Key"]).get("Value") != tag["Value"]
            ):
                return False

    return True



"""
Functions to sanitize the data in a present state
"""
import datetime
from typing import Any
from typing import Dict
from typing import List


def sanitize_dict(hub, data: Dict[str, Any]) -> Dict[str, Any]:
    """
    Iterate over a present state and verify that all the options are serializable
    """
    result = {}

    for key, value in data.items():
        if value is None:
            continue
        result[key] = hub.tool.aws.ec2.instance.data.sanitize_value(value)

    return result


def sanitize_list(hub, data: List[Any]) -> List[Any]:
    """
    Iterate over a present state and verify that all the options are serializable
    """
    result = []

    for value in data:
        if value is None:
            continue
        result.append(hub.tool.aws.ec2.instance.data.sanitize_value(value))

    return result


def sanitize_value(hub, data: Any) -> Any:
    """
    Iterate over a present state and verify that all the options are serializable
    """

    if isinstance(data, Dict):
        return hub.tool.aws.ec2.instance.data.sanitize_dict(data)
    elif isinstance(data, datetime.datetime):
        return str(data)
    elif isinstance(data, List):
        return hub.tool.aws.ec2.instance.data.sanitize_list(data)
    else:
        return data



"""State module for managing Lambda function asynchronous invocation configuration"""
import copy
from dataclasses import field
from dataclasses import make_dataclass
from typing import Any

__contracts__ = ["resource"]

from typing import Dict


async def present(
    hub,
    ctx,
    name: str,
    function_name: str,
    resource_id: str = None,
    qualifier: str = None,
    maximum_retry_attempts: int = None,
    maximum_event_age_in_seconds: int = None,
    destination_config: make_dataclass(
        """A destination configuration for events after they have been sent to a function."""
        "DestinationConfigRequest",
        [
            ("OnSuccess", Dict, field(default=None)),
            ("OnFailure", Dict, field(default=None)),
        ],
    ) = None,
) -> Dict[str, Any]:
    """
    Configures options for asynchronous invocation on a function, version, or alias. If a configuration already
    exists for a function, version, or alias, this operation overwrites it. By default, Lambda retries an asynchronous invocation twice if the function
    returns an error. It retains events in a queue for up to six hours. When an event fails all processing attempts
    or stays in the asynchronous invocation queue for too long, Lambda discards it. To retain discarded events,
    configure a dead-letter queue. To send an invocation record to a queue, topic,
    function, or event bus, specify a destination. You can configure separate destinations for successful
    invocations (on-success) and events that fail all processing attempts (on-failure). You can configure
    destinations in addition to or instead of a dead-letter queue.

    Args:
        name(str): An Idem name of the resource.
        resource_id(str, Optional): The Amazon Resource Name (ARN) of the function. Defaults to None.
        function_name(str):
            The name of the Lambda function. The length constraint applies only to the full ARN. If you  specify only the function name, it is limited to 64 characters in length.
                Name formats
                    * Function name - my-function
                    * Function ARN - ``arn:aws:lambda:us-west-2:123456789012:function:my-function``
                    * Partial ARN - ``123456789012:function:my-function``

        qualifier(str, Optional): A version number or alias name. Defaults to None.
        maximum_retry_attempts(int, Optional): The maximum number of times to retry when the function returns an error. Defaults to None.
        maximum_event_age_in_seconds(int, Optional): The maximum age of a request that Lambda sends to a function for processing. Defaults to None.
        destination_config(Dict, Optional):
            A destination for events after they have been sent to a function for processing.
                * (OnSuccess, Dict, Optional): The destination configuration for successful invocations.
                * (OnFailure, Dict, Optional): The destination configuration for failed invocations.
            Destinations
                * Function - The Amazon Resource Name (ARN) of a Lambda function.
                * Queue - The ARN of an SQS queue.
                * Topic - The ARN of an SNS topic.
                * Event Bus - The ARN of an Amazon EventBridge event bus. Defaults to None.

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: sls

            function_event_invoke_config-01234672f3336db8:
              aws.lambda_aws.function_event_invoke_config.present:
                - name: value
                - function_name: value
    """
    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)

    before = None
    resource_updated = False
    if resource_id:
        before = await hub.exec.aws.lambda_aws.function_event_invoke_config.get(
            ctx, name=name, resource_id=resource_id
        )
        if not (before["result"] and before["ret"]):
            result["comment"] = before["comment"]
            result["result"] = False
            return result
        result["old_state"] = copy.deepcopy(before["ret"])
        result["new_state"] = copy.deepcopy(result["old_state"])
        desired_resource_parameters = {
            "qualifier": qualifier,
            "maximum_retry_attempts": maximum_retry_attempts,
            "maximum_event_age_in_seconds": maximum_event_age_in_seconds,
            "destination_config": destination_config,
        }
        update_function_event_invoke_config_ret = await hub.tool.aws.lambda_aws.function_event_invoke_config.update_function_event_invoke_config(
            ctx,
            function_name=function_name,
            current_state=result["old_state"],
            desired_state=desired_resource_parameters,
        )
        if not update_function_event_invoke_config_ret["result"]:
            result["result"] = False
            result["comment"] = update_function_event_invoke_config_ret["comment"]
            return result

        result["comment"] = (
            result["comment"] + update_function_event_invoke_config_ret["comment"]
        )
        resource_updated = bool(update_function_event_invoke_config_ret["ret"])

        if ctx.get("test", False) and resource_updated:
            result["new_state"].update(update_function_event_invoke_config_ret["ret"])
            return result

    else:
        if ctx.get("test", False):
            result["new_state"] = hub.tool.aws.test_state_utils.generate_test_state(
                enforced_state={},
                desired_state={
                    "name": name,
                    "function_name": function_name,
                    "qualifier": qualifier,
                    "maximum_retry_attempts": maximum_retry_attempts,
                    "maximum_event_age_in_seconds": maximum_event_age_in_seconds,
                    "destination_config": destination_config,
                },
            )
            result["comment"] = hub.tool.aws.comment_utils.would_create_comment(
                resource_type="aws.lambda.function_event_invoke_config", name=name
            )
            return result

        create_ret = await hub.exec.boto3.client[
            "lambda"
        ].put_function_event_invoke_config(
            ctx,
            FunctionName=function_name,
            Qualifier=qualifier,
            MaximumRetryAttempts=maximum_retry_attempts,
            MaximumEventAgeInSeconds=maximum_event_age_in_seconds,
            DestinationConfig=destination_config,
        )
        result["result"] = create_ret["result"]
        if not result["result"]:
            result["comment"] = create_ret["comment"]
            return result
        result["comment"] = hub.tool.aws.comment_utils.create_comment(
            resource_type="aws.lambda.function_event_invoke_config", name=name
        )
    if (not before) or resource_updated:
        after = await hub.exec.aws.lambda_aws.function_event_invoke_config.get(
            ctx, name=name, function_name=function_name, qualifier=qualifier
        )
        if not (after["result"] and after["ret"]):
            result["result"] = False
            result["comment"] = result["comment"] + after["comment"]
            return result
        result["new_state"] = copy.deepcopy(after["ret"])

    return result


async def absent(
    hub,
    ctx,
    name: str,
    resource_id: str = None,
) -> Dict[str, Any]:
    """Deletes the configuration for asynchronous invocation for a function, version, or alias.

    Args:
        name(str): An Idem name of the resource.
        resource_id(str, Optional): The Amazon Resource Name (ARN) of the function. Idem automatically considers this
            resource being absent if this field is not specified.

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: sls

            function_event_invoke_config-01234672f3336db8:
              aws.lambda_aws.function_event_invoke_config.absent:
                - name: value
                - resource_id: value
    """
    already_absent_msg = hub.tool.aws.comment_utils.already_absent_comment(
        resource_type="aws.lambda.function_event_invoke_config", name=name
    )
    result = dict(
        comment=already_absent_msg,
        old_state=None,
        new_state=None,
        name=name,
        result=True,
    )

    if not resource_id:
        return result

    before = await hub.exec.aws.lambda_aws.function_event_invoke_config.get(
        ctx, name=name, resource_id=resource_id
    )

    if not before["result"]:
        result["comment"] = before["comment"]
        result["result"] = False
        return result

    if before["ret"]:
        result["old_state"] = before["ret"]
        if ctx.get("test", False):
            result["comment"] = hub.tool.aws.comment_utils.would_delete_comment(
                resource_type="aws.lambda.function_event_invoke_config", name=name
            )
        else:
            function_name = hub.tool.aws.arn_utils.get_resource_name(resource_id)
            qualifier = hub.tool.aws.arn_utils.get_qualifier(resource_id)
            ret = await hub.exec.boto3.client[
                "lambda"
            ].delete_function_event_invoke_config(
                ctx, **{"FunctionName": function_name, "Qualifier": qualifier}
            )
            result["result"] = ret["result"]
            if not result["result"]:
                result["comment"] = ret["comment"]
                return result
            result["comment"] = hub.tool.aws.comment_utils.delete_comment(
                resource_type="aws.lambda.function_event_invoke_config", name=name
            )

    return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    """Retrieves a list of configurations for asynchronous invocation for a function.

    Returns:
        Dict[str, Dict[str, Any]]

    Examples:
        .. code-block:: bash

            $ idem describe aws.lambda_aws.function_event_invoke_config
    """

    result = {}

    get_functions_ret = await hub.exec.boto3.client["lambda"].list_functions(ctx)
    if not get_functions_ret["result"]:
        hub.log.debug(f"Could not list functions {get_functions_ret['comment']}")
        return result

    for function in get_functions_ret["ret"]["Functions"]:
        function_name = function.get("FunctionName")
        get_event_invoke_configs_ret = await hub.exec.boto3.client[
            "lambda"
        ].list_function_event_invoke_configs(ctx, FunctionName=function_name)

        if not get_event_invoke_configs_ret["result"]:
            hub.log.debug(
                f"Could not get event invoke config for functionName '{function_name}': "
                f"{get_event_invoke_configs_ret['comment']}. Describe will skip this function and continue."
            )
            continue

        for function_event_invoke_config in get_event_invoke_configs_ret["ret"][
            "FunctionEventInvokeConfigs"
        ]:
            resource_arn = function_event_invoke_config.get("FunctionArn")
            # Including fields to match the 'present' function parameters
            resource_translated = hub.tool.aws.lambda_aws.function_event_invoke_config.convert_raw_function_event_invoke_config_to_present(
                function_event_invoke_config,
                function_name=function_name,
            )
            result[resource_arn] = {
                "aws.lambda.function_event_invoke_config.present": [
                    {parameter_key: parameter_value}
                    for parameter_key, parameter_value in resource_translated.items()
                ]
            }
    return result



from typing import List

# The instance must be stopped to modify this value
__update_state__ = "stopped"


async def apply(
    hub, ctx, resource, *, old_value: str, new_value: str, comments: List[str]
) -> bool:
    """
    Modify an ec2 instance based on a single parameter in it's "present" state

    Args:
        hub:
        ctx: The ctx from a state module call
        resource: An ec2 instance resource object
        old_value: The previous value from the attributes of an existing instance
        new_value: The desired value from the ec2 instance present state parameters
        comments: A running list of comments abound the update process
    """
    result = True
    # The instance must be stopped to modify its host resource group arn
    was_running = False
    if resource.state["Name"] == "running":
        was_running = True
        comments.append(f"Stopping the instance to prepare for changing kernel")
        ret = await hub.exec.boto3.client.ec2.stop_instances(
            InstanceIds=[resource.id], Hibernate=False, Force=False
        )
        await hub.tool.boto3.resource.exec(resource, "wait_until_stopped")

        if ret.comment:
            comments.append(ret.comment)
        if not ret.result:
            return False

    # Modify the host resource group arn
    ret = await hub.exec.boto3.client.ec2.modify_instance_placement(
        ctx,
        InstanceId=resource.id,
        HostResourceGroupArn=new_value,
        Affinity=resource.placement["Affinity"],
    )
    if ret.comment:
        comments.append(ret.comment)
    result &= ret.result

    if was_running:
        # Start the instance up again
        ret = await hub.exec.boto3.client.ec2.start_instances(InstanceIds=[resource.id])
        if ret.comment:
            comments.append(ret.comment)
        if not ret.result:
            return False
        await hub.tool.boto3.resource.exec(resource, "wait_until_running")

    return result



from typing import Any
from typing import Dict
from typing import List

from dict_tools import data


async def update(
    hub,
    ctx,
    resource_id: str,
    before: Dict[str, Any],
    bucket_logging_status: Dict[str, Any],
):
    """
    Updates the logging configuration for the S3 bucket.

    Args:
        hub: required for functions in hub.
        ctx: context.
        resource_id: AWS S3 bucket name.
        before(Dict): Existing resource parameters in Amazon Web Services.
        bucket_logging_status(Dict): Logging configuration.

    Returns:
        {"result": True|False, "comment": A message Tuple, "ret": Dict}

    """
    result = dict(comment=(), result=True, ret=None)

    # check and create payload for creating logging configurations to avoid redundant calls
    logging_diff = hub.tool.aws.s3.bucket_logging.get_logging_configuration_diff(
        bucket_logging_status, before
    )

    if logging_diff:
        result["ret"] = {}

        if not ctx.get("test", False):
            update_ret = await hub.exec.boto3.client.s3.put_bucket_logging(
                ctx=ctx, Bucket=resource_id, BucketLoggingStatus=bucket_logging_status
            )
            if not update_ret["result"]:
                result["comment"] = update_ret["comment"]
                result["result"] = False
                return result
            result["comment"] = hub.tool.aws.comment_utils.update_comment(
                resource_type="aws.s3.bucket_logging", name=resource_id
            )
        result["ret"]["bucket_logging_status"] = bucket_logging_status

    return result


def get_logging_configuration_diff(
    hub, new_logging_config: Dict[str, List], old_state: Dict[str, List]
) -> Dict[str, List]:
    """
    This functions helps in comparing two dicts.
    It compares each key value in both the dicts and return diff logging configuration

    Returns:
        {Resultant logging dictionary}

    """

    old_state_logging = old_state.get("bucket_logging_status")

    old_state_logging = old_state_logging if old_state_logging is not None else {}

    return data.recursive_diff(new_logging_config, old_state_logging, ignore_order=True)



"""State module for managing AWS Auto Scaling groups."""
import copy
from dataclasses import field
from dataclasses import make_dataclass
from typing import Any
from typing import Dict
from typing import List

__contracts__ = ["resource"]


async def present(
    hub,
    ctx,
    name: str,
    min_size: int,
    max_size: int,
    resource_id: str = None,
    launch_configuration_name: str = None,
    launch_template: make_dataclass(
        "LaunchTemplateSpecification",
        [
            ("LaunchTemplateId", str, field(default=None)),
            ("LaunchTemplateName", str, field(default=None)),
            ("Version", str, field(default=None)),
        ],
    ) = None,
    mixed_instances_policy: make_dataclass(
        "MixedInstancesPolicy",
        [
            (
                "LaunchTemplate",
                make_dataclass(
                    "LaunchTemplate",
                    [
                        (
                            "LaunchTemplateSpecification",
                            make_dataclass(
                                "LaunchTemplateSpecification",
                                [
                                    ("LaunchTemplateId", str, field(default=None)),
                                    ("LaunchTemplateName", str, field(default=None)),
                                    ("Version", str, field(default=None)),
                                ],
                            ),
                            field(default=None),
                        ),
                        (
                            "Overrides",
                            List[
                                make_dataclass(
                                    "LaunchTemplateOverrides",
                                    [
                                        ("InstanceType", str, field(default=None)),
                                        ("WeightedCapacity", str, field(default=None)),
                                        (
                                            "LaunchTemplateSpecification",
                                            make_dataclass(
                                                "LaunchTemplateSpecification",
                                                [
                                                    (
                                                        "LaunchTemplateId",
                                                        str,
                                                        field(default=None),
                                                    ),
                                                    (
                                                        "LaunchTemplateName",
                                                        str,
                                                        field(default=None),
                                                    ),
                                                    (
                                                        "Version",
                                                        str,
                                                        field(default=None),
                                                    ),
                                                ],
                                            ),
                                            field(default=None),
                                        ),
                                        (
                                            "InstanceRequirements",
                                            make_dataclass(
                                                "InstanceRequirements",
                                                [
                                                    (
                                                        "VCpuCount",
                                                        make_dataclass(
                                                            "VCpuCountRequest",
                                                            [
                                                                ("Min", int),
                                                                (
                                                                    "Max",
                                                                    int,
                                                                    field(default=None),
                                                                ),
                                                            ],
                                                        ),
                                                    ),
                                                    (
                                                        "MemoryMiB",
                                                        make_dataclass(
                                                            "MemoryMiBRequest",
                                                            [
                                                                ("Min", int),
                                                                (
                                                                    "Max",
                                                                    int,
                                                                    field(default=None),
                                                                ),
                                                            ],
                                                        ),
                                                    ),
                                                    (
                                                        "CpuManufacturers",
                                                        List[str],
                                                        field(default=None),
                                                    ),
                                                    (
                                                        "MemoryGiBPerVCpu",
                                                        make_dataclass(
                                                            "MemoryGiBPerVCpuRequest",
                                                            [
                                                                (
                                                                    "Min",
                                                                    float,
                                                                    field(default=None),
                                                                ),
                                                                (
                                                                    "Max",
                                                                    float,
                                                                    field(default=None),
                                                                ),
                                                            ],
                                                        ),
                                                        field(default=None),
                                                    ),
                                                    (
                                                        "ExcludedInstanceTypes",
                                                        List[str],
                                                        field(default=None),
                                                    ),
                                                    (
                                                        "InstanceGenerations",
                                                        List[str],
                                                        field(default=None),
                                                    ),
                                                    (
                                                        "SpotMaxPricePercentageOverLowestPrice",
                                                        int,
                                                        field(default=None),
                                                    ),
                                                    (
                                                        "OnDemandMaxPricePercentageOverLowestPrice",
                                                        int,
                                                        field(default=None),
                                                    ),
                                                    (
                                                        "BareMetal",
                                                        str,
                                                        field(default=None),
                                                    ),
                                                    (
                                                        "BurstablePerformance",
                                                        str,
                                                        field(default=None),
                                                    ),
                                                    (
                                                        "RequireHibernateSupport",
                                                        bool,
                                                        field(default=None),
                                                    ),
                                                    (
                                                        "NetworkInterfaceCount",
                                                        make_dataclass(
                                                            "NetworkInterfaceCountRequest",
                                                            [
                                                                (
                                                                    "Min",
                                                                    int,
                                                                    field(default=None),
                                                                ),
                                                                (
                                                                    "Max",
                                                                    int,
                                                                    field(default=None),
                                                                ),
                                                            ],
                                                        ),
                                                        field(default=None),
                                                    ),
                                                    (
                                                        "LocalStorage",
                                                        str,
                                                        field(default=None),
                                                    ),
                                                    (
                                                        "LocalStorageTypes",
                                                        List[str],
                                                        field(default=None),
                                                    ),
                                                    (
                                                        "TotalLocalStorageGB",
                                                        make_dataclass(
                                                            "TotalLocalStorageGBRequest",
                                                            [
                                                                (
                                                                    "Min",
                                                                    float,
                                                                    field(default=None),
                                                                ),
                                                                (
                                                                    "Max",
                                                                    float,
                                                                    field(default=None),
                                                                ),
                                                            ],
                                                        ),
                                                        field(default=None),
                                                    ),
                                                    (
                                                        "BaselineEbsBandwidthMbps",
                                                        make_dataclass(
                                                            "BaselineEbsBandwidthMbpsRequest",
                                                            [
                                                                (
                                                                    "Min",
                                                                    int,
                                                                    field(default=None),
                                                                ),
                                                                (
                                                                    "Max",
                                                                    int,
                                                                    field(default=None),
                                                                ),
                                                            ],
                                                        ),
                                                        field(default=None),
                                                    ),
                                                    (
                                                        "AcceleratorTypes",
                                                        List[str],
                                                        field(default=None),
                                                    ),
                                                    (
                                                        "AcceleratorCount",
                                                        make_dataclass(
                                                            "AcceleratorCountRequest",
                                                            [
                                                                (
                                                                    "Min",
                                                                    int,
                                                                    field(default=None),
                                                                ),
                                                                (
                                                                    "Max",
                                                                    int,
                                                                    field(default=None),
                                                                ),
                                                            ],
                                                        ),
                                                        field(default=None),
                                                    ),
                                                    (
                                                        "AcceleratorManufacturers",
                                                        List[str],
                                                        field(default=None),
                                                    ),
                                                    (
                                                        "AcceleratorNames",
                                                        List[str],
                                                        field(default=None),
                                                    ),
                                                    (
                                                        "AcceleratorTotalMemoryMiB",
                                                        make_dataclass(
                                                            "AcceleratorTotalMemoryMiBRequest",
                                                            [
                                                                (
                                                                    "Min",
                                                                    int,
                                                                    field(default=None),
                                                                ),
                                                                (
                                                                    "Max",
                                                                    int,
                                                                    field(default=None),
                                                                ),
                                                            ],
                                                        ),
                                                        field(default=None),
                                                    ),
                                                ],
                                            ),
                                            field(default=None),
                                        ),
                                    ],
                                )
                            ],
                            field(default=None),
                        ),
                    ],
                ),
                field(default=None),
            ),
            (
                "InstancesDistribution",
                make_dataclass(
                    "InstancesDistribution",
                    [
                        ("OnDemandAllocationStrategy", str, field(default=None)),
                        ("OnDemandBaseCapacity", int, field(default=None)),
                        (
                            "OnDemandPercentageAboveBaseCapacity",
                            int,
                            field(default=None),
                        ),
                        ("SpotAllocationStrategy", str, field(default=None)),
                        ("SpotInstancePools", int, field(default=None)),
                        ("SpotMaxPrice", str, field(default=None)),
                    ],
                ),
                field(default=None),
            ),
        ],
    ) = None,
    instance_id: str = None,
    desired_capacity: int = None,
    default_cooldown: int = None,
    availability_zones: List[str] = None,
    load_balancer_names: List[str] = None,
    target_group_ar_ns: List[str] = None,
    health_check_type: str = None,
    health_check_grace_period: int = None,
    placement_group: str = None,
    vpc_zone_identifier: str = None,
    termination_policies: List[str] = None,
    new_instances_protected_from_scale_in: bool = None,
    capacity_rebalance: bool = None,
    lifecycle_hook_specification_list: List[
        make_dataclass(
            "LifecycleHookSpecification",
            [
                ("LifecycleHookName", str, field(default=None)),
                ("LifecycleTransition", str, field(default=None)),
                ("NotificationMetadata", str, field(default=None)),
                ("HeartbeatTimeout", int, field(default=None)),
                ("DefaultResult", str, field(default=None)),
                ("NotificationTargetARN", str, field(default=None)),
                ("RoleARN", str, field(default=None)),
            ],
        )
    ] = None,
    tags: Dict[str, Any]
    or List[
        make_dataclass(
            "Tag",
            [
                ("key", str, field(default=None)),
                ("value", str, field(default=None)),
                ("propagate_at_launch", bool, field(default=None)),
            ],
        )
    ] = None,
    service_linked_role_arn: str = None,
    max_instance_lifetime: int = None,
    desired_capacity_type: str = None,
) -> Dict[str, Any]:
    r"""Creates an Auto Scaling group with the specified name and attributes.

    We strongly recommend using a launch template when calling this operation to ensure full functionality for Amazon
    EC2 Auto Scaling and Amazon EC2.

    If you exceed your maximum limit of Auto Scaling groups, the call fails. To query this limit, call the
    DescribeAccountLimits API. For information about updating this limit, see Amazon EC2 Auto Scaling service quotas in
    the Amazon EC2 Auto Scaling User Guide.

    For introductory exercises for creating an Auto Scaling group, see Getting started with Amazon EC2 Auto Scaling and
    Tutorial: Set up a scaled and load-balanced application in the Amazon EC2 Auto Scaling User Guide.

    For more information, see Auto Scaling groups in the Amazon EC2 Auto Scaling User Guide. Every Auto Scaling group
    has three size parameters (`DesiredCapacity`, `MaxSize`, and `MinSize`). Usually, you set these sizes based on a
    specific number of instances. However, if you configure a mixed instances policy that defines weights for the
    instance types, you must specify these sizes with the same units that you use for weighting instances.

    Args:
        resource_id(str, Optional):
            The name of the Auto Scaling group. This name must be unique per Region per account.

        name(str):
            An Idem name of the resource. This is also used as the AutoScaling Group Name during resource creation.

        launch_configuration_name(str, Optional):
            The name of the launch configuration to use to launch instances.

            Conditional: You must specify either a launch template (launch_template or mixed_instances_policy) or a
            launch configuration (launch_configuration_name or instance_id).

            Defaults to None.

        launch_template(dict[str, Any], Optional):
            Information used to specify the launch template and version to use to launch instances.

            Conditional: You must specify either a launch template (launch_template or mixed_instances_policy) or a
            launch configuration (launch_configuration_name or instance_id).

            .. Note::
                The launch template that is specified must be configured for use with an Auto Scaling group. For more
                information, see Creating a launch template for an Auto Scaling group in the Amazon EC2 Auto Scaling
                User Guide.

            Defaults to None.

            * LaunchTemplateId (str, Optional):
                The ID of the launch template. To get the template ID, use the Amazon EC2
                DescribeLaunchTemplates API operation. New launch templates can be created using the Amazon EC2
                CreateLaunchTemplate API.

                Conditional: You must specify either a LaunchTemplateId or a LaunchTemplateName.

            * LaunchTemplateName (str, Optional):
                The name of the launch template. To get the template name, use the Amazon EC2 DescribeLaunchTemplates
                API operation. New launch templates can be created using the Amazon EC2 CreateLaunchTemplate API.

                Conditional: You must specify either a LaunchTemplateId or a LaunchTemplateName.

            * Version (str, Optional):
                The version number, `$Latest`, or `$Default`. To get the version number, use the Amazon EC2
                DescribeLaunchTemplateVersions API operation. New launch template versions can be created using
                the Amazon EC2 CreateLaunchTemplateVersion API. If the value is `$Latest`, Amazon EC2 Auto Scaling
                selects the latest version of the launch template when launching instances. If the value is
                `$Default`, Amazon EC2 Auto Scaling selects the default version of the launch template when
                launching instances. The default value is `$Default`.

        mixed_instances_policy(dict[str, Any], Optional):
            An embedded object that specifies a mixed instances policy.

            For more information, see Auto Scaling groups with multiple instance types and purchase options in the
            Amazon EC2 Auto Scaling User Guide.

            Defaults to None.

            * LaunchTemplate (dict[str, Any], Optional):
                Specifies the launch template to use and the instance types (overrides) that are used to launch EC2
                instances to fulfill On-Demand and Spot capacities.

                Required when creating a mixed instances policy.

                * LaunchTemplateSpecification (dict[str, Any], Optional):
                    The launch template to use.

                    * LaunchTemplateId (str, Optional):
                        The ID of the launch template. To get the template ID, use the Amazon EC2
                        DescribeLaunchTemplates API operation. New launch templates can be created using the Amazon EC2
                        CreateLaunchTemplate API.

                        Conditional: You must specify either a LaunchTemplateId or a LaunchTemplateName.

                    * LaunchTemplateName (str, Optional):
                        The name of the launch template. To get the template name, use the Amazon EC2
                        DescribeLaunchTemplates API operation. New launch templates can be created using the Amazon EC2
                        CreateLaunchTemplate API.

                        Conditional: You must specify either a LaunchTemplateId or a LaunchTemplateName.

                    * Version (str, Optional):
                        The version number, `$Latest`, or `$Default`. To get the version number, use the Amazon EC2
                        DescribeLaunchTemplateVersions API operation. New launch template versions can be created using
                        the Amazon EC2 CreateLaunchTemplateVersion API. If the value is `$Latest`, Amazon EC2 Auto
                        Scaling selects the latest version of the launch template when launching instances. If the
                        value is `$Default`, Amazon EC2 Auto Scaling selects the default version of the launch template
                        when launching instances. The default value is `$Default`.

                * Overrides (list[dict[str, Any]], Optional):
                    Any properties that you specify override the same properties in the launch template. If not
                    provided, Amazon EC2 Auto Scaling uses the instance type or instance type requirements specified
                    in the launch template when it launches an instance.

                    The overrides can include either one or more instance types or a set of instance requirements, but
                    not both.

                    * InstanceType (str, Optional):
                        The instance type, such as m3.xlarge. You must use an instance type that is supported in your
                        requested Region and Availability Zones. For more information, see Instance types in the Amazon
                        Elastic Compute Cloud User Guide.

                    * WeightedCapacity (str, Optional):
                        The number of capacity units provided by the instance type specified in InstanceType in terms of
                        virtual CPUs, memory, storage, throughput, or other relative performance characteristic. When a
                        Spot or On-Demand Instance is launched, the capacity units count toward the desired capacity.
                        Amazon EC2 Auto Scaling launches instances until the desired capacity is totally fulfilled, even
                        if this results in an overage. For example, if there are two units remaining to fulfill
                        capacity, and Amazon EC2 Auto Scaling can only launch an instance with a WeightedCapacity of
                        five units, the instance is launched, and the desired capacity is exceeded by three units. For
                        more information, see Configuring instance weighting for Amazon EC2 Auto Scaling in the Amazon
                        EC2 Auto Scaling User Guide. Value must be in the range of 1–999.

                    * LaunchTemplateSpecification (dict[str, Any], Optional):
                        Provides a launch template for the specified instance type or instance requirements. For
                        example, some instance types might require a launch template with a different AMI. If not
                        provided, Amazon EC2 Auto Scaling uses the launch template that's defined for your mixed
                        instances policy. For more information, see Specifying a different launch template for an
                        instance type in the Amazon EC2 Auto Scaling User Guide.

                        * LaunchTemplateId (str, Optional):
                            The ID of the launch template. To get the template ID, use the Amazon EC2
                            DescribeLaunchTemplates API operation. New launch templates can be created using the Amazon
                            EC2 CreateLaunchTemplate API.

                            Conditional: You must specify either a LaunchTemplateId or a LaunchTemplateName.

                        * LaunchTemplateName (str, Optional):
                            The name of the launch template. To get the template name, use the Amazon EC2
                            DescribeLaunchTemplates API operation. New launch templates can be created using the Amazon
                            EC2 CreateLaunchTemplate API.

                            Conditional: You must specify either a LaunchTemplateId or a LaunchTemplateName.

                        * Version (str, Optional):
                            The version number, `$Latest`, or `$Default`. To get the version number, use the Amazon EC2
                            DescribeLaunchTemplateVersions API operation. New launch template versions can be created
                            using the Amazon EC2 CreateLaunchTemplateVersion API. If the value is `$Latest`, Amazon EC2
                            Auto Scaling selects the latest version of the launch template when launching instances. If
                            the value is `$Default`, Amazon EC2 Auto Scaling selects the default version of the launch
                            template when launching instances. The default value is `$Default`.

                    * InstanceRequirements (dict[str, Any], Optional):
                        The instance requirements. When you specify instance requirements, Amazon EC2 Auto Scaling finds
                        instance types that satisfy your requirements, and then uses your On-Demand and Spot allocation
                        strategies to launch instances from these instance types, in the same way as when you specify a
                        list of specific instance types.

                        * VCpuCount (dict[str, Any]):
                            The minimum and maximum number of vCPUs for an instance type.

                            * Min (int): The minimum number of vCPUs.
                            * Max (int, Optional): The maximum number of vCPUs.

                        * MemoryMiB (dict[str, Any]):
                            The minimum and maximum instance memory size for an instance type, in MiB.

                            * Min (int): The memory minimum in MiB.
                            * Max (int, Optional): The memory maximum in MiB.

                        * CpuManufacturers (list[str], Optional):
                            Lists which specific CPU manufacturers to include.

                            * For instance types with Intel CPUs, specify intel.
                            * For instance types with AMD CPUs, specify amd.
                            * For instance types with Amazon Web Services CPUs, specify amazon-web-services.

                            .. Note::
                                Don't confuse the CPU hardware manufacturer with the CPU hardware architecture.
                                Instances will be launched with a compatible CPU architecture based on the Amazon
                                Machine Image (AMI) that you specify in your launch template.

                            Default: Any manufacturer

                        * MemoryGiBPerVCpu (dict[str, Any], Optional):
                            The minimum and maximum amount of memory per vCPU for an instance type, in GiB.

                            Default: No minimum or maximum

                            * Min (float, Optional): The memory minimum in GiB.
                            * Max (float, Optional): The memory maximum in GiB.

                        * ExcludedInstanceTypes (list[str], Optional):
                            Lists which instance types to exclude. You can use strings with one or more wild cards,
                            represented by an asterisk (*). The following are examples: c5*, m5a.*, r*, *3*.

                            For example, if you specify c5*, you are excluding the entire C5 instance family, which
                            includes all C5a and C5n instance types. If you specify m5a.*, you are excluding all the
                            M5a instance types, but not the M5n instance types.

                            Default: No excluded instance types

                        * InstanceGenerations (list[str], Optional):
                            Indicates whether current or previous generation instance types are included.

                            * For current generation instance types, specify current.
                            The current generation includes EC2 instance types currently recommended for use. This
                            typically includes the latest two to three generations in each instance family.
                            For more information, see Instance types in the Amazon EC2 User Guide for Linux Instances.

                            * For previous generation instance types, specify previous.

                            Default: Any current or previous generation

                        * SpotMaxPricePercentageOverLowestPrice (int, Optional):
                            The price protection threshold for Spot Instances. This is the maximum you’ll pay for a Spot
                            Instance, expressed as a percentage higher than the cheapest M, C, or R instance type with
                            your specified attributes. When Amazon EC2 Auto Scaling selects instance types with your
                            attributes, we will exclude instance types whose price is higher than your threshold. The
                            parameter accepts an integer, which Amazon EC2 Auto Scaling interprets as a percentage. To
                            turn off price protection, specify a high value, such as 999999.

                            If you set DesiredCapacityType to vcpu or memory-mib, the price protection threshold is
                            applied based on the per vCPU or per memory price instead of the per instance price.

                            Default: 100

                        * OnDemandMaxPricePercentageOverLowestPrice (int, Optional):
                            The price protection threshold for On-Demand Instances. This is the maximum you’ll pay for
                            an On-Demand Instance, expressed as a percentage higher than the cheapest M, C, or R
                            instance type with your specified attributes. When Amazon EC2 Auto Scaling selects instance
                            types with your attributes, we will exclude instance types whose price is higher than your
                            threshold. The parameter accepts an integer, which Amazon EC2 Auto Scaling interprets as a
                            percentage. To turn off price protection, specify a high value, such as 999999.

                            If you set DesiredCapacityType to vcpu or memory-mib, the price protection threshold is
                            applied based on the per vCPU or per memory price instead of the per instance price.

                            Default: 20

                        * BareMetal (str, Optional):
                            Indicates whether bare metal instance types are included, excluded, or required.

                            Default: excluded

                        * BurstablePerformance (str, Optional):
                            Indicates whether burstable performance instance types are included, excluded, or required.
                            For more information, see Burstable performance instances in the Amazon EC2 User Guide for
                            Linux Instances.

                            Default: excluded

                        * RequireHibernateSupport (bool, Optional):
                            Indicates whether instance types must provide On-Demand Instance hibernation support.

                            Default: false

                        * NetworkInterfaceCount (dict[str, Any], Optional):
                            The minimum and maximum number of network interfaces for an instance type.

                            Default: No minimum or maximum

                            * Min (int, Optional): The minimum number of network interfaces.
                            * Max (int, Optional): The maximum number of network interfaces.

                        * LocalStorage (str, Optional):
                            Indicates whether instance types with instance store volumes are included, excluded, or
                            required. For more information, see Amazon EC2 instance store in the Amazon EC2 User Guide
                            for Linux Instances.

                            Default: included

                        * LocalStorageTypes (list[str], Optional):
                            Indicates the type of local storage that is required.

                            * For instance types with hard disk drive (HDD) storage, specify hdd.
                            * For instance types with solid state drive (SSD) storage, specify sdd.

                            Default: Any local storage type

                        * TotalLocalStorageGB (dict[str, Any], Optional):
                            The minimum and maximum total local storage size for an instance type, in GB.

                            Default: No minimum or maximum

                            * Min (float, Optional): The storage minimum in GB.
                            * Max (float, Optional): The storage maximum in GB.

                        * BaselineEbsBandwidthMbps (dict[str, Any], Optional):
                            The minimum and maximum baseline bandwidth performance for an instance type, in Mbps. For
                            more information, see Amazon EBS–optimized instances in the Amazon EC2 User Guide for Linux
                            Instances.

                            Default: No minimum or maximum

                            * Min (int, Optional): The minimum value in Mbps.
                            * Max (int, Optional): The maximum value in Mbps.

                        * AcceleratorTypes (list[str], Optional):
                            Lists the accelerator types that must be on an instance type.

                            * For instance types with GPU accelerators, specify gpu.
                            * For instance types with FPGA accelerators, specify fpga.
                            * For instance types with inference accelerators, specify inference.

                            Default: Any accelerator type

                        * AcceleratorCount (dict[str, Any], Optional):
                            The minimum and maximum number of accelerators (GPUs, FPGAs, or Amazon Web Services
                            Inferentia chips) for an instance type.

                            To exclude accelerator-enabled instance types, set Max to 0.

                            Default: No minimum or maximum

                            * Min (int, Optional): The minimum value.
                            * Max (int, Optional): The maximum value.

                        * AcceleratorManufacturers (list[str], Optional):
                            Indicates whether instance types must have accelerators by specific manufacturers.

                            * For instance types with NVIDIA devices, specify nvidia.
                            * For instance types with AMD devices, specify amd.
                            * For instance types with Amazon Web Services devices, specify amazon-web-services.
                            * For instance types with Xilinx devices, specify xilinx.

                            Default: Any manufacturer

                        * AcceleratorNames (list[str], Optional):
                            Lists the accelerators that must be on an instance type.

                            * For instance types with NVIDIA A100 GPUs, specify a100.
                            * For instance types with NVIDIA V100 GPUs, specify v100.
                            * For instance types with NVIDIA K80 GPUs, specify k80.
                            * For instance types with NVIDIA T4 GPUs, specify t4.
                            * For instance types with NVIDIA M60 GPUs, specify m60.
                            * For instance types with AMD Radeon Pro V520 GPUs, specify radeon-pro-v520.
                            * For instance types with Xilinx VU9P FPGAs, specify vu9p.

                            Default: Any accelerator

                        * AcceleratorTotalMemoryMiB (dict[str, Any], Optional):
                            The minimum and maximum total memory size for the accelerators on an instance type, in MiB.

                            Default: No minimum or maximum

                            * Min (int, Optional): The memory minimum in MiB.
                            * Max (int, Optional): The memory maximum in MiB.

            * InstancesDistribution (dict[str, Any], Optional):
                Specifies the instances distribution.

                * OnDemandAllocationStrategy (str, Optional):
                    The order of the launch template overrides to use in fulfilling On-Demand capacity.

                    If you specify lowest-price, Amazon EC2 Auto Scaling uses price to determine the order, launching
                    the lowest price first.

                    If you specify prioritized, Amazon EC2 Auto Scaling uses the priority that you assigned to each
                    launch template override, launching the highest priority first. If all your On-Demand capacity
                    cannot be fulfilled using your highest priority instance, then Amazon EC2 Auto Scaling launches the
                    remaining capacity using the second priority instance type, and so on.

                    Default: lowest-price for Auto Scaling groups that specify InstanceRequirements in the overrides
                    and prioritized for Auto Scaling groups that don't.

                    Valid values: lowest-price | prioritized

                * OnDemandBaseCapacity (int, Optional):
                    The minimum amount of the Auto Scaling group's capacity that must be fulfilled by On-Demand
                    Instances. This base portion is launched first as your group scales.

                    If you specify weights for the instance types in the overrides, the base capacity is measured in
                    the same unit of measurement as the instance types. If you specify InstanceRequirements in the
                    overrides, the base capacity is measured in the same unit of measurement as your group's desired
                    capacity.

                    Default: 0

                * OnDemandPercentageAboveBaseCapacity (int, Optional):
                    Controls the percentages of On-Demand Instances and Spot Instances for your additional capacity
                    beyond OnDemandBaseCapacity. Expressed as a number (for example, 20 specifies 20% On-Demand
                    Instances, 80% Spot Instances). If set to 100, only On-Demand Instances are used.

                    Default: 100

                * SpotAllocationStrategy (str, Optional):
                    Indicates how to allocate instances across Spot Instance pools.

                    If the allocation strategy is lowest-price, the Auto Scaling group launches instances using the
                    Spot pools with the lowest price, and evenly allocates your instances across the number of Spot
                    pools that you specify.

                    If the allocation strategy is capacity-optimized (recommended), the Auto Scaling group launches
                    instances using Spot pools that are optimally chosen based on the available Spot capacity.
                    Alternatively, you can use capacity-optimized-prioritized and set the order of instance types in
                    the list of launch template overrides from highest to lowest priority (from first to last in the
                    list). Amazon EC2 Auto Scaling honors the instance type priorities on a best-effort basis but
                    optimizes for capacity first.

                    Default: lowest-price

                    Valid values: lowest-price | capacity-optimized | capacity-optimized-prioritized

                * SpotInstancePools (int, Optional):
                    The number of Spot Instance pools across which to allocate your Spot Instances. The Spot pools
                    are determined from the different instance types in the overrides. Valid only when the Spot
                    allocation strategy is lowest-price. Value must be in the range of 1–20.

                    Default: 2

                * SpotMaxPrice (str, Optional):
                    The maximum price per unit hour that you are willing to pay for a Spot Instance. If you keep the
                    value at its default (unspecified), Amazon EC2 Auto Scaling uses the On-Demand price as the
                    maximum Spot price. To remove a value that you previously set, include the property but specify
                    an empty string ("") for the value.

                    .. Warning::
                        If your maximum price is lower than the Spot price for the instance types that you selected,
                        your Spot Instances are not launched.

                    Valid Range: Minimum value of 0.001

        instance_id(str, Optional):
            The ID of the instance used to base the launch configuration on. If specified, Amazon EC2 Auto
            Scaling uses the configuration values from the specified instance to create a new launch configuration. To
            get the instance ID, use the Amazon EC2 DescribeInstances API operation. For more information, see Creating
            an Auto Scaling group using an EC2 instance in the Amazon EC2 Auto Scaling User Guide.

            Defaults to None.

        min_size(int):
            The minimum size of the group.

        max_size(int):
            The maximum size of the group.

            .. Note::
                With a mixed instances policy that uses instance weighting, Amazon EC2 Auto Scaling may need to go
                above MaxSize to meet your capacity requirements. In this event, Amazon EC2 Auto Scaling will never go
                above MaxSize by more than your largest instance weight (weights that define how many units each
                instance contributes to the desired capacity of the group).

        desired_capacity(int, Optional):
            The desired capacity is the initial capacity of the Auto Scaling group at the time of its
            creation and the capacity it attempts to maintain. It can scale beyond this capacity if you
            configure auto scaling. This number must be greater than or equal to the minimum size of the
            group and less than or equal to the maximum size of the group. If you do not specify a desired
            capacity, the default is the minimum size of the group. Defaults to None.

        default_cooldown(int, Optional):
            The amount of time, in seconds, after a scaling activity completes before another scaling
            activity can start. The default value is 300. This setting applies when using simple scaling
            policies, but not when using other scaling policies or scheduled scaling. For more information,
            see Scaling cooldowns for Amazon EC2 Auto Scaling in the Amazon EC2 Auto Scaling User Guide.

            Default: 300 seconds

        availability_zones(list, Optional):
            A list of Availability Zones where instances in the Auto Scaling group can be created. Used for launching
            into the default VPC subnet in each Availability Zone when not using the VPCZoneIdentifier property, or for
            attaching a network interface when an existing network interface ID is specified in a launch template.
            Defaults to None.

        load_balancer_names(list, Optional):
            A list of Classic Load Balancers associated with this Auto Scaling group. For Application Load
            Balancers, Network Load Balancers, and Gateway Load Balancers, specify the TargetGroupARNs
            property instead. Defaults to None.

        target_group_ar_ns(list, Optional):
            The Amazon Resource Names (ARN) of the target groups to associate with the Auto Scaling group.
            Instances are registered as targets in a target group, and traffic is routed to the target
            group. For more information, see Elastic Load Balancing and Amazon EC2 Auto Scaling in the
            Amazon EC2 Auto Scaling User Guide. Defaults to None.

        health_check_type(str, Optional):
            The service to use for the health checks. The valid values are EC2 (default) and ELB. If you
            configure an Auto Scaling group to use load balancer (ELB) health checks, it considers the
            instance unhealthy if it fails either the EC2 status checks or the load balancer health checks.
            For more information, see Health checks for Auto Scaling instances in the Amazon EC2 Auto
            Scaling User Guide. Defaults to None.

        health_check_grace_period(int, Optional):
            The amount of time, in seconds, that Amazon EC2 Auto Scaling waits before checking the health
            status of an EC2 instance that has come into service and marking it unhealthy due to a failed
            health check. The default value is 0. For more information, see Health check grace period in the
            Amazon EC2 Auto Scaling User Guide. Required if you are adding an ELB health check.

            Default: 0 seconds

        placement_group(str, Optional):
            The name of the placement group into which to launch your instances. For more information,
            see Placement Groups in the Amazon EC2 User Guide for Linux Instances.

            .. Note::
                A placement group is a logical grouping of instances within a single Availability Zone. You cannot
                specify multiple Availability Zones and a placement group.

            Defaults to None.

        vpc_zone_identifier(str, Optional):
            A comma-separated list of subnet IDs for a virtual private cloud (VPC) where instances in the
            Auto Scaling group can be created. If you specify VPCZoneIdentifier with AvailabilityZones, the
            subnets that you specify must reside in those Availability Zones.

        termination_policies(list[str], Optional):
            A policy or a list of policies that are used to select the instance to terminate. These policies
            are executed in the order that you list them. For more information, see Controlling which Auto
            Scaling instances terminate during scale in the Amazon EC2 Auto Scaling User Guide. Defaults to None.

            Valid values: Default | AllocationStrategy | ClosestToNextInstanceHour | NewestInstance
            | OldestInstance | OldestLaunchConfiguration | OldestLaunchTemplate
            | arn:aws:lambda:region:account-id:function:my-function:my-alias

        new_instances_protected_from_scale_in(bool, Optional):
            Indicates whether newly launched instances are protected from termination by Amazon EC2 Auto
            Scaling when scaling in. For more information about preventing instances from terminating on
            scale in, see Using instance scale-in protection in the Amazon EC2 Auto Scaling User Guide.
            Defaults to None.

        capacity_rebalance(bool, Optional):
            Indicates whether Capacity Rebalancing is enabled. Otherwise, Capacity Rebalancing is disabled.
            When you turn on Capacity Rebalancing, Amazon EC2 Auto Scaling attempts to launch a Spot
            Instance whenever Amazon EC2 notifies that a Spot Instance is at an elevated risk of
            interruption. After launching a new instance, it then terminates an old instance. For more
            information, see Amazon EC2 Auto Scaling Capacity Rebalancing in the Amazon EC2 Auto Scaling
            User Guide. Defaults to None.

        lifecycle_hook_specification_list(list[dict[str, Any]], Optional):
            One or more lifecycle hooks for the group, which specify actions to perform when Amazon EC2 Auto
            Scaling launches or terminates instances. Defaults to None.

            * LifecycleHookName (str):
                The name of the lifecycle hook.

            * LifecycleTransition (str):
                The state of the EC2 instance to which you want to attach the lifecycle hook. For Auto Scaling groups,
                there are two major lifecycle transitions.

                * To create a lifecycle hook for scale-out events, specify autoscaling:EC2_INSTANCE_LAUNCHING.
                * To create a lifecycle hook for scale-in events, specify autoscaling:EC2_INSTANCE_TERMINATING.

            * NotificationMetadata (str, Optional):
                Additional information that you want to include any time Amazon EC2 Auto Scaling sends a message
                to the notification target.

            * HeartbeatTimeout (int, Optional):
                The maximum time, in seconds, that can elapse before the lifecycle hook times out.
                The range is from 30 to 7200 seconds. The default value is 3600 seconds (1 hour).

            * DefaultResult (str, Optional):
                Defines the action the Auto Scaling group should take when the lifecycle hook timeout elapses or
                if an unexpected failure occurs. The valid values are CONTINUE and ABANDON. The default value is
                ABANDON.

            * NotificationTargetARN (str, Optional):
                The ARN of the target that Amazon EC2 Auto Scaling sends notifications to when an instance is in
                the transition state for the lifecycle hook. The notification target can be either an SQS queue
                or an SNS topic.

            * RoleARN (str, Optional):
                The ARN of the IAM role that allows the Auto Scaling group to publish to the specified
                notification target. Valid only if the notification target is an Amazon SNS topic or an Amazon
                SQS queue. Required for new lifecycle hooks, but optional when updating existing hooks.

        tags(dict or list, Optional):
            Dict in the format of {tag-key: tag-value, propagate_at_launch-{tag-key}: value}
            or list of tags in the format of [{"key": tag-key, "value": tag-value, "propagate_at_launch": value}].

            One or more tags. You can tag your Auto Scaling group and propagate the tags to the Amazon EC2
            instances it launches. Tags are not propagated to Amazon EBS volumes. To add tags to Amazon EBS
            volumes, specify the tags in a launch template but use caution. If the launch template specifies
            an instance tag with a key that is also specified for the Auto Scaling group, Amazon EC2 Auto
            Scaling overrides the value of that instance tag with the value specified by the Auto Scaling
            group. For more information, see Tagging Auto Scaling groups and instances in the Amazon EC2
            Auto Scaling User Guide. Defaults to None.

            * (key):
                The tag key.
            * (value, Optional):
                The tag value.
            * (propagate_at_launch, Optional):
                Determines whether the tag is added to new instances as they are launched in the group.

        service_linked_role_arn(str, Optional):
            The Amazon Resource Name (ARN) of the service-linked role that the Auto Scaling group uses to
            call other Amazon Web Services on your behalf. By default, Amazon EC2 Auto Scaling uses a
            service-linked role named AWSServiceRoleForAutoScaling, which it creates if it does not exist.
            For more information, see Service-linked roles in the Amazon EC2 Auto Scaling User Guide. Defaults to None.

        max_instance_lifetime(int, Optional):
            The maximum amount of time, in seconds, that an instance can be in service. The default is null.
            If specified, the value must be either 0 or a number equal to or greater than 86,400 seconds (1
            day). For more information, see Replacing Auto Scaling instances based on maximum instance
            lifetime in the Amazon EC2 Auto Scaling User Guide. Defaults to None.

        desired_capacity_type(str, Optional):
            The unit of measurement for the value specified for desired capacity. Amazon EC2 Auto Scaling
            supports DesiredCapacityType for attribute-based instance type selection only. For more
            information, see Creating an Auto Scaling group using attribute-based instance type selection in
            the Amazon EC2 Auto Scaling User Guide.

            By default, Amazon EC2 Auto Scaling specifies units, which translates into number of instances.

            Valid values: units | vcpu | memory-mib. Defaults to None.

    Request Syntax:
        .. code-block:: sls

            [auto-scaling-group-resource-id]:
              aws.autoscaling.auto_scaling_group.present:
                - resource_id: "string"
                - name: "string"
                - launch_configuration_name: "string"
                - min_size: "int"
                - max_size: "int"
                - desired_capacity: "int"
                - default_cooldown: "int"
                - availability_zones: List
                - health_check_type: "string"
                - health_check_grace_period: "int"
                - vpc_zone_identifier: "string"
                - termination_policies: List
                - new_instances_protected_from_scale_in: "bool"
                - tags: Dict or List


    Returns:
        Dict[str, Any]

    Examples:
        .. code-block:: sls

            idem-scaling-group:
              aws.autoscaling.auto_scaling_group.present:
                - resource_id: idem-scaling-group
                - name: idem-scaling-group
                - launch_configuration_name: idem-launch_configuration
                - min_size: 2
                - max_size: 4
                - desired_capacity: 2
                - default_cooldown: 300
                - availability_zones:
                    - us-west-1b
                    - us-west-1c
                - health_check_type: EC2
                - health_check_grace_period: 300
                - vpc_zone_identifier: subnet-xxxxxxx,subnet-xxxxxxx,subnet-xxxxxx
                - termination_policies:
                    - Default
                - new_instances_protected_from_scale_in: false
                - tags(tags in list format):
                    - key: Identifier
                      value: idem-aws
                      propagate_at_launch: true
                - key: Name
                  propagate_at_launch: false
                  value: idem-scaling-group
              - tags(or in dict format):
                  Identifier: idem-aws
                  propagate_at_launch-Identifier: true
                  Name: idem-scaling-group
                  propagate_at_launch-Name: false
    """
    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)

    # Query for auto scaling groups
    auto_scaling_group = None
    if resource_id:
        auto_scaling_group = (
            await hub.exec.boto3.client.autoscaling.describe_auto_scaling_groups(
                ctx, AutoScalingGroupNames=[resource_id]
            )
        )
        if not auto_scaling_group["result"]:
            result["comment"] = auto_scaling_group["comment"]
            result["result"] = auto_scaling_group["result"]
            return result
    if isinstance(tags, List):
        tags = hub.tool.aws.tag_utils.convert_auto_scaling_tag_list_to_dict(tags)

    # Check for existing auto scaling group
    before = None
    if auto_scaling_group and auto_scaling_group["ret"]["AutoScalingGroups"]:
        before = auto_scaling_group["ret"]["AutoScalingGroups"][0]

    # Update current state
    current_state = hub.tool.aws.autoscaling.auto_scaling_group_utils.convert_raw_auto_scaling_group_to_present(
        auto_scaling_group=before
    )
    result["old_state"] = current_state

    # Handle No change behaviour
    desired_state = {
        "name": name,
        "resource_id": resource_id,
        "launch_configuration_name": launch_configuration_name,
        "launch_template": launch_template,
        "instance_id": instance_id,
        "min_size": min_size,
        "max_size": max_size,
        "desired_capacity": desired_capacity,
        "default_cooldown": default_cooldown,
        "availability_zones": availability_zones,
        "load_balancer_names": load_balancer_names,
        "target_group_ar_ns": target_group_ar_ns,
        "health_check_type": health_check_type,
        "health_check_grace_period": health_check_grace_period,
        "placement_group": placement_group,
        "mixed_instances_policy": mixed_instances_policy,
        "vpc_zone_identifier": vpc_zone_identifier,
        "termination_policies": termination_policies,
        "new_instances_protected_from_scale_in": new_instances_protected_from_scale_in,
        "capacity_rebalance": capacity_rebalance,
        "lifecycle_hook_specification_list": lifecycle_hook_specification_list,
        "tags": tags,
        "max_instance_lifetime": max_instance_lifetime,
        "desired_capacity_type": desired_capacity_type,
    }

    is_auto_scaling_group_updated = (
        hub.tool.aws.autoscaling.auto_scaling_group_utils.is_auto_scaling_group_updated(
            current_state=current_state, desired_state=desired_state
        )
    )

    is_tags_updated = current_state.get("tags") != desired_state.get("tags")

    is_change_detected = (
        before is None or is_auto_scaling_group_updated or is_tags_updated
    )

    if not is_change_detected:
        result["comment"] = (
            f"aws.autoscaling.auto_scaling_group '{name}' already exists",
        )
        result["new_state"] = copy.deepcopy(result["old_state"])
        return result

    # Handle test behaviour
    if ctx.get("test", False):
        result["new_state"] = hub.tool.aws.test_state_utils.generate_test_state(
            enforced_state=current_state,
            desired_state=desired_state,
        )
        result["comment"] = (
            hub.tool.aws.comment_utils.would_update_comment(
                resource_type="aws.autoscaling.auto_scaling_group", name=name
            )
            if before
            else hub.tool.aws.comment_utils.would_create_comment(
                resource_type="aws.autoscaling.auto_scaling_group", name=name
            )
        )
        return result

    # Handle actual resource create or update
    if before:
        # Update auto scaling group
        if is_auto_scaling_group_updated:
            ret = await hub.exec.boto3.client.autoscaling.update_auto_scaling_group(
                ctx,
                **{
                    "AutoScalingGroupName": name,
                    "LaunchConfigurationName": launch_configuration_name,
                    "LaunchTemplate": launch_template,
                    "MixedInstancesPolicy": mixed_instances_policy,
                    "InstanceId": instance_id,
                    "MinSize": min_size,
                    "MaxSize": max_size,
                    "DesiredCapacity": desired_capacity,
                    "DefaultCooldown": default_cooldown,
                    "AvailabilityZones": availability_zones,
                    "HealthCheckType": health_check_type,
                    "HealthCheckGracePeriod": health_check_grace_period,
                    "PlacementGroup": placement_group,
                    "VPCZoneIdentifier": vpc_zone_identifier,
                    "TerminationPolicies": termination_policies,
                    "NewInstancesProtectedFromScaleIn": new_instances_protected_from_scale_in,
                    "CapacityRebalance": capacity_rebalance,
                    "LifecycleHookSpecificationList": lifecycle_hook_specification_list,
                    "ServiceLinkedRoleARN": service_linked_role_arn,
                    "MaxInstanceLifetime": max_instance_lifetime,
                    "DesiredCapacityType": desired_capacity_type,
                },
            )
            result["result"] = ret["result"]
            if not result["result"]:
                result["comment"] = ret["comment"]
                return result
            result["comment"] = hub.tool.aws.comment_utils.update_comment(
                resource_type="aws.autoscaling.auto_scaling_group", name=name
            )

        # Update tags
        if is_tags_updated:
            update_tags = await hub.tool.aws.autoscaling.tag.update_tags(
                ctx,
                old_tags=current_state["tags"],
                new_tags=desired_state["tags"],
                resource_id=resource_id,
            )
            result["comment"] = result["comment"] + update_tags["comment"]
            result["result"] = result["result"] and update_tags["result"]
            if not result["result"]:
                result["comment"] = update_tags["comment"]
                return result
    else:
        # Handle actual resource creation
        tags_list = hub.tool.aws.tag_utils.convert_auto_scaling_tag_dict_to_list(tags)
        translated_tags = hub.tool.aws.autoscaling.auto_scaling_group_utils.convert_present_tags_to_raw_tags(
            resource_id=name, tags=tags_list
        )
        ret = await hub.exec.boto3.client.autoscaling.create_auto_scaling_group(
            ctx,
            **{
                "AutoScalingGroupName": name,
                "LaunchConfigurationName": launch_configuration_name,
                "LaunchTemplate": launch_template,
                "MixedInstancesPolicy": mixed_instances_policy,
                "InstanceId": instance_id,
                "MinSize": min_size,
                "MaxSize": max_size,
                "DesiredCapacity": desired_capacity,
                "DefaultCooldown": default_cooldown,
                "AvailabilityZones": availability_zones,
                "LoadBalancerNames": load_balancer_names,
                "TargetGroupARNs": target_group_ar_ns,
                "HealthCheckType": health_check_type,
                "HealthCheckGracePeriod": health_check_grace_period,
                "PlacementGroup": placement_group,
                "VPCZoneIdentifier": vpc_zone_identifier,
                "TerminationPolicies": termination_policies,
                "NewInstancesProtectedFromScaleIn": new_instances_protected_from_scale_in,
                "CapacityRebalance": capacity_rebalance,
                "LifecycleHookSpecificationList": lifecycle_hook_specification_list,
                "Tags": translated_tags,
                "ServiceLinkedRoleARN": service_linked_role_arn,
                "MaxInstanceLifetime": max_instance_lifetime,
                "DesiredCapacityType": desired_capacity_type,
            },
        )
        result["result"] = ret["result"]
        if not result["result"]:
            result["comment"] = ret["comment"]
            return result
        result["comment"] = hub.tool.aws.comment_utils.create_comment(
            resource_type="aws.autoscaling.auto_scaling_group", name=name
        )
        resource_id = name

    # Populate new state
    new_auto_scaling_groups = (
        await hub.exec.boto3.client.autoscaling.describe_auto_scaling_groups(
            ctx, AutoScalingGroupNames=[resource_id]
        )
    )
    if not new_auto_scaling_groups["result"]:
        result["comment"] = result["comment"] + new_auto_scaling_groups["comment"]
        result["result"] = new_auto_scaling_groups["result"]
        return result

    after = new_auto_scaling_groups["ret"]["AutoScalingGroups"][0]
    result[
        "new_state"
    ] = hub.tool.aws.autoscaling.auto_scaling_group_utils.convert_raw_auto_scaling_group_to_present(
        auto_scaling_group=after
    )
    return result


async def absent(
    hub,
    ctx,
    name: str,
    resource_id: str = None,
    timeout: Dict = None,
) -> Dict[str, Any]:
    r"""Deletes the specified Auto Scaling group.

    If the group has instances or scaling activities in progress, you must specify the option to force the deletion
    in order for it to succeed.

    If the group has policies, deleting the group deletes the policies, the underlying alarm actions, and any alarm
    that no longer has an associated action.

    To remove instances from the Auto Scaling group before deleting it, call the DetachInstances API with the list
    of instances and the option to decrement the desired capacity. This ensures that Amazon EC2 Auto Scaling does
    not launch replacement instances.

    To terminate all instances before deleting the Auto Scaling group, call the UpdateAutoScalingGroup API and set the
    minimum size and desired capacity of the Auto Scaling group to zero.

    Args:
        resource_id(str, Optional):
            The name of the Auto Scaling group.
        name(str):
            An Idem name of the resource.
        timeout(dict, Optional):
            Timeout configuration for deletion of AWS auto scaling group.

            * delete (dict) -- Timeout configuration for deletion of an auto scaling group.
                * delay -- The amount of time in seconds to wait between attempts. Defaults to 15
                * max_attempts -- Customized timeout configuration containing delay and max attempts. Defaults to 40

    Request Syntax:
        .. code-block:: sls

            [auto-scaling-group-resource-id]:
              aws.autoscaling.auto_scaling_group.absent:
                - resource_id: "string"
                - name: "string"

    Returns:
        Dict[str, Any]

    Examples:
        .. code-block:: sls

            idem-scaling-group:
              aws.autoscaling.auto_scaling_group.absent:
                - resource_id: idem-scaling-group
                - name: idem-scaling-group
    """
    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)

    # Query for auto scaling groups
    auto_scaling_group = None
    if resource_id:
        auto_scaling_group = (
            await hub.exec.boto3.client.autoscaling.describe_auto_scaling_groups(
                ctx, AutoScalingGroupNames=[resource_id]
            )
        )

    # Check for existing auto scaling group
    before = None
    if (
        auto_scaling_group
        and auto_scaling_group["result"]
        and auto_scaling_group["ret"]["AutoScalingGroups"]
    ):
        before = auto_scaling_group["ret"]["AutoScalingGroups"][0]
        result[
            "old_state"
        ] = hub.tool.aws.autoscaling.auto_scaling_group_utils.convert_raw_auto_scaling_group_to_present(
            auto_scaling_group=before
        )

    if not before:
        result["comment"] = hub.tool.aws.comment_utils.already_absent_comment(
            resource_type="aws.autoscaling.auto_scaling_group", name=name
        )

    elif ctx.get("test", False):
        result["comment"] = hub.tool.aws.comment_utils.would_delete_comment(
            resource_type="aws.autoscaling.auto_scaling_group", name=name
        )

    else:
        ret = await hub.exec.boto3.client.autoscaling.delete_auto_scaling_group(
            ctx,
            AutoScalingGroupName=resource_id,
            ForceDelete=True,
        )
        if not ret["result"]:
            result["result"] = ret["result"]
            result["comment"] = ret["comment"]
            return result

        # Custom waiter for delete
        waiter_config = hub.tool.aws.waiter_utils.create_waiter_config(
            default_delay=15,
            default_max_attempts=40,
            timeout_config=timeout.get("delete") if timeout else None,
        )
        delete_waiter_acceptors = [
            {
                "matcher": "path",
                "expected": True,
                "state": "success",
                "argument": "length(AutoScalingGroups) == `0`",
            },
            {
                "matcher": "path",
                "expected": True,
                "state": "retry",
                "argument": "length(AutoScalingGroups) > `0`",
            },
        ]
        auto_scaling_group_waiter = hub.tool.boto3.custom_waiter.waiter_wrapper(
            name="AutoScalingGroupDelete",
            operation="DescribeAutoScalingGroups",
            argument=["AutoScalingGroups"],
            acceptors=delete_waiter_acceptors,
            client=hub.tool.boto3.client.get_client(ctx, "autoscaling"),
        )
        try:
            await hub.tool.boto3.client.wait(
                ctx,
                "autoscaling",
                "AutoScalingGroupDelete",
                auto_scaling_group_waiter,
                AutoScalingGroupNames=[
                    resource_id,
                ],
                WaiterConfig=waiter_config,
            )
        except Exception as e:
            result["comment"] = result["comment"] + (str(e),)
            result["result"] = False
            return result

        result["comment"] = hub.tool.aws.comment_utils.delete_comment(
            resource_type="aws.autoscaling.auto_scaling_group", name=name
        )

    return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    r"""Gets information about the Auto Scaling groups in the account and Region.

    Describe the resource in a way that can be recreated/managed with the corresponding "present" function.


    Returns:
        Dict[str, Dict[str, Any]]

    Examples:
        .. code-block:: bash

            $ idem describe aws.autoscaling.auto_scaling_group
    """
    result = {}
    ret = await hub.exec.boto3.client.autoscaling.describe_auto_scaling_groups(ctx)
    if not ret["result"]:
        hub.log.debug(f"Could not describe auto_scaling_group {ret['comment']}")
        return {}

    for auto_scaling_group in ret["ret"]["AutoScalingGroups"]:
        resource_id = auto_scaling_group["AutoScalingGroupName"]
        resource = hub.tool.aws.autoscaling.auto_scaling_group_utils.convert_raw_auto_scaling_group_to_present(
            auto_scaling_group=auto_scaling_group
        )
        result[resource_id] = {
            "aws.autoscaling.auto_scaling_group.present": [
                {parameter_key: parameter_value}
                for parameter_key, parameter_value in resource.items()
            ]
        }
    return result



"""State module for managing Amazon Organizations."""
import copy
from typing import Any
from typing import Dict

__contracts__ = ["resource"]

AWS_Org_Not_In_Use_Exception = "AWSOrganizationsNotInUseException"
ALREADY_EXISTS = "already exists"

TREQ = {
    "absent": {
        "require": [
            "aws.organizations.organization_unit.absent",
        ],
    },
}


async def present(hub, ctx, name: str, feature_set: str) -> Dict[str, Any]:
    """Creates an AWS organization.

    The account whose user is calling the operation automatically becomes the management account of the new organization.
    This operation must be called using credentials from the account that is to become the new organization's management account.
    The principal must also have the relevant IAM permissions.

    Args:
        name(str):
            The name of the organization.
        feature_set(str):
            Specifies the feature set supported by the new organization. Each feature set supports different levels of functionality.

            * ``CONSOLIDATED_BILLING``: All member accounts have their bills consolidated to and paid by the management account.
            * ``ALL``: In addition to all the features supported by the consolidated billing feature set, the management account can also apply any policy type to any member account in the organization.

    Request Syntax:
      .. code-block:: sls

        [idem_test_aws_organizations_organization]:
          aws.organizations.organization.present:
          - name: 'string'
          - feature_set: 'string'

    Returns:
        Dict[str, Any]

    Examples:
        .. code-block:: sls

            idem_test_aws_organizations_organization:
              aws.organizations.organization.present:
                - name: 'idem_test_organization'
                - feature_set: 'ALL
    """
    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)

    before = await hub.exec.boto3.client.organizations.describe_organization(ctx)

    if not before:
        # Org not present , create

        if ctx.get("test", False):
            result["new_state"] = hub.tool.aws.test_state_utils.generate_test_state(
                enforced_state={},
                desired_state={
                    "name": name,
                    "feature_set": feature_set,
                },
            )
            result["comment"] = result["comment"] + (
                f"Would create aws.organizations.organization '{name}'.",
            )
            return result

        try:
            ret = await hub.exec.boto3.client.organizations.create_organization(
                ctx, FeatureSet=feature_set
            )

            result["result"] = ret["result"]
            if not result["result"]:
                result["comment"] = result["comment"] + ret["comment"]
                return result
            result["comment"] = result["comment"] + (
                f"Created aws.organizations.organization {name}.",
            )
            after = await hub.exec.boto3.client.organizations.describe_organization(ctx)
            if after and after.get("ret"):
                result[
                    "new_state"
                ] = hub.tool.aws.organizations.conversion_utils.convert_raw_organization_to_present(
                    after["ret"]["Organization"], name
                )
                roots_resp = await hub.exec.boto3.client.organizations.list_roots(ctx)

                if roots_resp and roots_resp["ret"] and roots_resp["ret"]["Roots"]:
                    result["new_state"]["roots"] = roots_resp["ret"]["Roots"]
        except hub.tool.boto3.exception.ClientError as e:
            result["comment"] = result["comment"] + (f"{e.__class__.__name__}: {e}",)
            result["result"] = False
    else:
        # Org present
        result[
            "old_state"
        ] = hub.tool.aws.organizations.conversion_utils.convert_raw_organization_to_present(
            before["ret"]["Organization"], name
        )
        roots_resp = await hub.exec.boto3.client.organizations.list_roots(ctx)

        if roots_resp and roots_resp["ret"] and roots_resp["ret"]["Roots"]:
            result["old_state"]["roots"] = roots_resp["ret"]["Roots"]
        result["new_state"] = copy.deepcopy(result["old_state"])
        result["comment"] = result["comment"] + (
            f"aws.organizations.organization '{name}' {ALREADY_EXISTS}.",
        )

    return result


async def absent(hub, ctx, name: str) -> Dict[str, Any]:
    """Deletes the organization.

    You can delete an organization only by using credentials from the management account. The organization must be empty of member accounts.

    Args:
        name(str):
            The name of the organization.

    Request syntax:
      .. code-block:: sls

        [idem_test_aws_organizations_organization]:
          aws.organizations.organization.absent:
            - name: 'string'

    Returns:
        Dict[str, Any]

    Examples:
        .. code-block:: sls

            idem_test_aws_organizations_organization:
              aws.organizations.organization.absent:
                - name: 'idem_test_organization'
    """
    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)

    before = await hub.exec.boto3.client.organizations.describe_organization(ctx)
    if before:
        result[
            "old_state"
        ] = hub.tool.aws.organizations.conversion_utils.convert_raw_organization_to_present(
            before["ret"], name
        )
        if ctx.get("test", False):
            result["comment"] = result["comment"] + (
                f"Would delete aws.organizations.organization '{name}'.",
            )
            return result

    if not before and AWS_Org_Not_In_Use_Exception in str(before["comment"]):
        result["comment"] = result["comment"] + (
            f"aws.organizations.organization '{name}' is already absent",
        )
        return result
    else:
        try:
            ret = await hub.exec.boto3.client.organizations.delete_organization(ctx)

            if not ret["result"]:
                result["comment"] = ret["comment"]
                result["result"] = False
                return result
            result["comment"] = result["comment"] + (
                f"aws.organizations.organization {name} deleted.",
            )
            result["result"] = ret["result"]
        except hub.tool.boto3.exception.ClientError as e:
            result["comment"] = (f"{e.__class__.__name__}: {e}",)

    try:
        after = await hub.exec.boto3.client.organizations.describe_organization(ctx)
        if after.get("ret"):
            result[
                "new_state"
            ] = hub.tool.aws.organizations.conversion_utils.convert_raw_organization_to_present(
                after["ret"], name
            )
    except Exception as e:
        result["comment"] = result["comment"] + (str(e),)
        result["result"] = False
    return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    """Describes AWS Organizations in a way that can be recreated/managed with the corresponding "present" function.

    Returns:
        Dict[str, Dict[str, Any]

    Examples:
        .. code-block:: bash

            $ idem describe aws.organizations.organization
    """
    result = {}
    ret = await hub.exec.boto3.client.organizations.describe_organization(ctx)
    if not ret["result"]:
        hub.log.debug(f"Could not describe organization with error:  {ret['comment']}")
        return {}

    resource = ret["ret"]["Organization"]

    translated_resource = (
        hub.tool.aws.organizations.conversion_utils.convert_raw_organization_to_present(
            resource
        )
    )
    roots_resp = await hub.exec.boto3.client.organizations.list_roots(ctx)

    if roots_resp and roots_resp["ret"] and roots_resp["ret"]["Roots"]:
        translated_resource["roots"] = roots_resp["ret"]["Roots"]
    result[resource["Id"]] = {
        "aws.organizations.organization.present": [
            {parameter_key: parameter_value}
            for parameter_key, parameter_value in translated_resource.items()
        ]
    }

    return result



def pre(hub, ctx):
    """
    Before using esm, verify that a region_name is part of the profile
    """
    kwargs = ctx.get_arguments()
    func_ctx = kwargs["ctx"]

    if func_ctx.acct.get("region_name"):
        # Region name already exists for the profile
        return

    # If there is no region_name, get it from hub.OPT
    opt = getattr(hub, "OPT") or {}
    acct = opt.get("acct") or {}
    extras = acct.get("extras") or {}
    aws_opts = extras.get("aws") or {}
    esm_opts = aws_opts.get("esm") or {}

    # Get the region name from the esm config
    if esm_opts.get("region_name"):
        hub.log.debug(
            "Getting region_name from hub.OPT.acct.extras.aws.esm.region_name"
        )
        func_ctx.acct["region_name"] = esm_opts.get("region_name")

    elif aws_opts.get("region_name"):
        # Fall back to the general config for region_name in acct.extras.aws.region_name
        hub.log.debug("Getting region_name from hub.OPT.acct.extras.aws.region_name")
        func_ctx.acct["region_name"] = aws_opts.get("region_name")



from collections import OrderedDict
from typing import Any
from typing import Dict
from typing import List

from dict_tools import differ


async def search_raw(
    hub,
    ctx,
    resource_id: str = None,
    load_balancer_arn: str = None,
) -> Dict:
    """
    1. Describes specified listeners or the listeners for the specified Application Load Balancer,
        Network Load Balancer or Gateway Load Balancer.
    2. You must specify either a load balancer or one or more listeners. The return will be in the same
        format as what boto3 api returns.
    3. Here, resource_id get higher priority in search than load_balancer_arn i.e. if both load_balancer_arn
        and resource_id are not None, search is done with resource_id than load_balancer_arn.

    Args:
        resource_id(str, optional):
            AWS ELBv2 Listener ARN to identify the resource.

        load_balancer_arn(str, optional):
            The Amazon Resource Name (ARN) of the load balancer.

    Returns:
        {"result": True|False, "comment": A message List, "ret": Dict}

    """
    ret = result = dict(comment=[], ret=None, result=True)
    if resource_id:
        ret = await hub.exec.boto3.client.elbv2.describe_listeners(
            ctx,
            ListenerArns=[resource_id],
        )
    elif load_balancer_arn:
        ret = await hub.exec.boto3.client.elbv2.describe_listeners(
            ctx,
            LoadBalancerArn=load_balancer_arn,
        )
    result["result"] = ret["result"]
    result["comment"] = list(ret["comment"])
    result["ret"] = ret["ret"]
    return result


async def update(
    hub,
    ctx,
    name: str,
    current_state: Dict[str, Any],
    input_map: Dict[str, Any],
    resource_id: str,
    plan_state: Dict[str, Any],
):
    r"""
    1. Replaces the specified properties of the specified listener. Any property that is not updated, remains unchanged.
    2. Changing the protocol from HTTPS to HTTP, or from TLS to TCP, removes the security policy and default certificate
       properties. If you change the protocol from HTTP to HTTPS, or from TCP to TLS, you must add the security policy
       and default certificate properties.
    3. To add an item to a list, remove an item from a list, or update an item in a list, you must provide the entire
       list. For example, to add an action, specify a list with the current actions plus the new action.
    4. Adds the specified SSL server certificate to the certificate list for the specified HTTPS/ TLS listener. If the
       certificate in already in the certificate list, the call is successful but the certificate is not added again.
    5. Removes the specified certificate from the certificate list for the specified HTTPS or TLS listener.

    Args:
        name(str):
            The name of the AWS ElasticLoadBalancingv2 Listener.

        current_state(dict[str, Any]):
            response returned by describe on an AWS ElasticLoadBalancingv2 Listener

        input_map(dict[str, Any]):
            a dictionary with newly passed values of params.

        resource_id(str):
            AWS ElasticLoadBalancingv2 Listener Amazon Resource Name (ARN).

        plan_state(dict[str, Any]):
            idem --test state for update on AWS ElasticLoadBalancingv2 Listener.

    Returns:
        {"result": True|False, "comment": A message List, "ret": None}
    """
    result = dict(comment=[], result=True, ret=[])
    if not ctx.get("test", False):
        if input_map:
            ret = compare_config(current_state=current_state, updated_state=input_map)
            if (ret["result"] and ret["ret"]) or input_map.get("default_actions"):
                modify_listener = await hub.exec.boto3.client.elbv2.modify_listener(
                    ctx,
                    ListenerArn=resource_id,
                    Port=ret["ret"].get("Port"),
                    Protocol=ret["ret"].get("Protocol"),
                    SslPolicy=ret["ret"].get("SslPolicy"),
                    Certificates=ret["ret"].get("DefaultCertificates"),
                    DefaultActions=input_map.get("default_actions"),
                    AlpnPolicy=ret["ret"].get("AlpnPolicy"),
                )
                if not modify_listener["result"]:
                    result["comment"] = list(modify_listener["comment"])
                    result["result"] = False
                    return result
                result["comment"].append("Modified Listener.")
                result["ret"].append(
                    {"modify_listener": modify_listener["ret"]["Listeners"][0]}
                )

        if input_map.get("certificates"):
            ret = compare_certificates(
                old_certificates=current_state.get("certificates"),
                new_certificates=input_map.get("certificates"),
            )
            if ret["ret"]:
                if ret["ret"].get("to_remove"):
                    remove_certificates = (
                        await hub.exec.boto3.client.elbv2.remove_listener_certificates(
                            ctx,
                            ListenerArn=resource_id,
                            Certificates=ret["ret"]["to_remove"],
                        )
                    )
                    if not remove_certificates["result"]:
                        result["comment"] += list(remove_certificates["comment"])
                        result["result"] = False
                        return result
                    result["comment"].append("Removed Certificates.")
                    result["ret"].append(
                        {"remove_certificates": remove_certificates["ret"]}
                    )

                if ret["ret"].get("to_add"):
                    add_certificates = (
                        await hub.exec.boto3.client.elbv2.add_listener_certificates(
                            ctx,
                            ListenerArn=resource_id,
                            Certificates=ret["ret"]["to_add"],
                        )
                    )
                    if not add_certificates["result"]:
                        result["comment"] += list(add_certificates["comment"])
                        result["result"] = False
                        return result
                    result["comment"].append("Added Certificates.")
                    result["ret"].append({"add_certificates": add_certificates["ret"]})
    else:
        update_params = OrderedDict(
            {
                "name": name,
                "default_actions": input_map.get("default_actions"),
                "protocol": input_map.get("protocol"),
                "port": input_map.get("port"),
                "ssl_policy": input_map.get("ssl_policy"),
                "certificates": input_map.get("certificates"),
                "default_certificates": input_map.get("default_certificates"),
                "alpn_policy": input_map.get("alpn_policy"),
                "tags": input_map.get("tags"),
                "resource_id": resource_id,
            }
        )
        for key, value in update_params.items():
            if value is not None:
                plan_state[key] = value
        result["ret"] = plan_state
    return result


def compare_certificates(
    old_certificates: List[Dict[str, Any]] = None,
    new_certificates: List[Dict[str, Any]] = None,
):
    """
    Compares old_certificates and new_certificates and return the new list of certificates that need to be updated.

    Args:
        old_certificates(list[dict[str, Any]]):
            Existing list of certificates to be removed from AWS ElasticLoadBalancingv2 Listener.

        new_certificates(list[dict[str, Any]]):
            Newer list of certificates to be added to AWS ElasticLoadBalancingv2 Listener.

    Returns:
        {"comment": ("A tuple",), "ret": Dict}

    """
    result = dict(ret=None)
    to_remove = []
    to_add = []
    old_certificates_map = {
        certificate.get("CertificateArn"): certificate
        for certificate in old_certificates or []
    }
    if new_certificates is not None:
        for certificate in new_certificates:
            if certificate.get("CertificateArn") in old_certificates_map:
                del old_certificates_map[certificate.get("CertificateArn")]
            else:
                to_add.append(certificate)
        to_remove = list(old_certificates_map.values())
        result["ret"] = {"to_add": to_add, "to_remove": to_remove}
    return result


def compare_config(current_state: Dict[str, Any], updated_state: Dict[str, Any]):
    """
    Compares listener's existing configuration with newly passed inputs and returns details of the diff.

    Args:
        current_state (dict[str, Any]):
            response returned by describe on an AWS ELBv2 Listener

        updated_state (dict[st, Any]):
            Newly passed values for update function as dictionary .

    Returns:
        {"result": True|False, "ret": Dict}

    """
    result = dict(result=False, ret={})
    target_group_params = {
        "Port": "port",
        "Protocol": "protocol",
        "SslPolicy": "ssl_policy",
        "AlpnPolicy": "alpn_policy",
        "DefaultCertificates": "default_certificates",
    }
    current_config_mapping = {}
    for name, value in target_group_params.items():
        if current_state.get(value):
            current_config_mapping[name] = current_state.get(value)

    updated_config_mapping = {}
    for name, value in target_group_params.items():
        if updated_state.get(value):
            updated_config_mapping[name] = updated_state.get(value)

    diff_in_config = differ.deep_diff(current_config_mapping, updated_config_mapping)

    if diff_in_config.get("new"):
        result["result"] = True
        result["ret"] = diff_in_config.get("new")

    return result



"""
Autogenerated using `pop-create-idem <https://gitlab.com/saltstack/pop/pop-create-idem>`__

hub.exec.boto3.client.eks.create_addon
hub.exec.boto3.client.eks.delete_addon
hub.exec.boto3.client.eks.describe_addon
hub.exec.boto3.client.eks.list_addons
hub.exec.boto3.client.eks.update_addon
"""
import copy
from dataclasses import field
from dataclasses import make_dataclass
from typing import Any
from typing import Dict

__contracts__ = ["resource"]

TREQ = {
    "absent": {
        "require": [
            "aws.ec2.subnet.absent",
            "aws.iam.role.absent",
            "aws.iam.role_policy_attachment.absent",
            "aws.eks.cluster.absent",
        ],
    },
    "present": {
        "require": [
            "aws.ec2.subnet.present",
            "aws.iam.role.present",
            "aws.iam.role_policy_attachment.present",
            "aws.eks.cluster.present",
        ],
    },
}


async def present(
    hub,
    ctx,
    name: str,
    cluster_name: str,
    addon_version: str,
    resource_id: str = None,
    service_account_role_arn: str = None,
    resolve_conflicts: str = "OVERWRITE",
    tags: Dict[str, str] = None,
    timeout: make_dataclass(
        "Timeout",
        [
            (
                "create",
                make_dataclass(
                    "CreateTimeout",
                    [
                        ("delay", int, field(default=10)),
                        ("max_attempts", int, field(default=60)),
                    ],
                ),
                field(default=None),
            ),
            (
                "update",
                make_dataclass(
                    "UpdateTimeout",
                    [
                        ("delay", int, field(default=10)),
                        ("max_attempts", int, field(default=60)),
                    ],
                ),
                field(default=None),
            ),
        ],
    ) = None,
) -> Dict[str, Any]:
    """
    **Autogenerated function**

    Creates an Amazon EKS add-on. Amazon EKS add-ons help to automate the provisioning and lifecycle management of
    common operational software for Amazon EKS clusters. Amazon EKS add-ons require clusters running version 1.18 or
    later because Amazon EKS add-ons rely on the Server-side Apply Kubernetes feature, which is only available in
    Kubernetes 1.18 and later. For more information, see Amazon EKS add-ons in the Amazon EKS User Guide.

    Args:
        name(Text): An Idem name of the EKS addon resource.
        resource_id(Text, optional): AWS EKS addon name.
        cluster_name(Text): The name of the cluster to create the add-on for
        addon_version(Text): The version of the add-on. The version must match one of the versions returned
            by ` DescribeAddonVersions
        service_account_role_arn(Text, optional):The Amazon Resource Name (ARN) of an existing IAM role to bind to the add-on's
            service account. The role must be assigned the IAM permissions required by the add-on. If you don't specify
            an existing IAM role, then the add-on uses the permissions assigned to the node IAM role.
        resolve_conflicts(Text, optional): How to resolve parameter value conflicts when migrating an existing add-on
            to an Amazon EKS add-on
        tags(Dict, optional): The metadata to apply to the cluster to assist with categorization and organization.
            Each tag consists of a key and an optional value. You define both
        timeout(Dict, optional): Timeout configuration for creating or updating addon.
            * create (Dict) -- Timeout configuration for creating addon
                * delay(int, default=10) -- The amount of time in seconds to wait between attempts.
                * max_attempts(int, default=60) -- Customized timeout configuration containing delay and max attempts.
            * update (string) -- Timeout configuration for updating addon
                * delay(int, default=10) -- The amount of time in seconds to wait between attempts.
                * max_attempts(int, default=60) -- Customized timeout configuration containing delay and max attempts.
    Request Syntax:

        [eks-addon-name]:
              aws.eks.addon.present:
              - cluster_name: 'string'
              - addon_version: 'string'
              - resource_id: 'string'
              - service_account_role_arn: 'string'
              - resolve_conflicts: 'string'
              - tags:
                - Key: 'string'
                  Value: 'string'
              - timeout:
                   create:
                      delay: 'integer'
                      max_attempts: 'integer'
                   update:
                      delay: 'integer'
                      max_attempts: 'integer

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: sls

            kube-proxy:
              aws.eks.addon.present:
              - cluster_name: eks-12j44i4k
              - addon_version: v1.19.8-eksbuild.1
              - resource_id: kube-proxy
              - service_account_role_arn: arn:role-ejwew124
              - resolve_conflicts: "OVERWRITE"
              - tags:
                - Key: Name
                  Value: eks-addon-name
    """

    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)
    before = None
    resource_updated = False
    if resource_id:
        before = await hub.exec.boto3.client.eks.describe_addon(
            ctx, clusterName=cluster_name, addonName=name
        )
        if not before["result"]:
            if not ("ResourceNotFoundException" in before["comment"][0]):
                result["result"] = False
                result["comment"] = before["comment"]
                return result

    if before and before["ret"]:
        try:
            result[
                "old_state"
            ] = hub.tool.aws.eks.conversion_utils.convert_raw_addon_to_present(
                raw_resource=before["ret"]["addon"], idem_resource_name=name
            )
            plan_state = copy.deepcopy(result["old_state"])
            client_request_token = result["old_state"].get("client_request_token", None)

            update_ret = await hub.exec.aws.eks.addon.update_addon(
                ctx=ctx,
                name=name,
                before=result["old_state"],
                addon_version=addon_version,
                service_account_role_arn=service_account_role_arn,
                resolve_conflicts=resolve_conflicts,
                client_request_token=client_request_token,
                timeout=timeout,
            )
            result["comment"] = update_ret["comment"]
            result["result"] = update_ret["result"]
            resource_updated = bool(update_ret["ret"])

            if update_ret["ret"] and ctx.get("test", False):
                result["comment"] = result["comment"] + (
                    f"Would update aws.eks.addon '{name}'",
                )
                for addon_param in update_ret["ret"]:
                    plan_state[addon_param] = update_ret["ret"][addon_param]

            if tags is not None:
                # Update tags
                update_tags_ret = await hub.exec.aws.eks.tag.update_eks_tags(
                    ctx=ctx,
                    resource_arn=result["old_state"].get("addon_arn"),
                    old_tags=result["old_state"].get("tags"),
                    new_tags=tags,
                )

            result["comment"] = result["comment"] + update_tags_ret["comment"]
            result["result"] = result["result"] and update_tags_ret["result"]
            resource_updated = resource_updated or bool(update_tags_ret["ret"])
            if ctx.get("test", False) and update_tags_ret["ret"]:
                plan_state["tags"] = update_tags_ret["ret"].get("tags")
        except hub.tool.boto3.exception.ClientError as e:
            result["comment"] = result["comment"] + (f"{e.__class__.__name__}: {e}",)
            result["result"] = False
    else:
        if ctx.get("test", False):
            result["new_state"] = hub.tool.aws.test_state_utils.generate_test_state(
                enforced_state={},
                desired_state={
                    "name": name,
                    "cluster_name": cluster_name,
                    "addon_version": addon_version,
                    "service_account_role_arn": service_account_role_arn,
                    "resolve_conflicts": resolve_conflicts,
                    "tags": tags,
                },
            )
            result["comment"] = hub.tool.aws.comment_utils.would_create_comment(
                resource_type="aws.eks.addon", name=name
            )
            return result
        try:
            ret = await hub.exec.boto3.client.eks.create_addon(
                ctx,
                clusterName=cluster_name,
                addonName=name,
                addonVersion=addon_version,
                serviceAccountRoleArn=service_account_role_arn,
                resolveConflicts=resolve_conflicts,
                tags=tags,
            )
            result["result"] = ret["result"]
            if not result["result"]:
                result["comment"] = ret["comment"]
                return result

            resource_id = ret["ret"]["addon"]["addonName"]
            waiter_config = hub.tool.aws.waiter_utils.create_waiter_config(
                default_delay=60,
                default_max_attempts=40,
                timeout_config=timeout.get("create") if timeout else None,
            )
            hub.log.debug(f"Waiting on creating aws.eks.addon '{name}'")
            try:
                await hub.tool.boto3.client.wait(
                    ctx,
                    "eks",
                    "addon_active",
                    addonName=resource_id,
                    clusterName=cluster_name,
                    WaiterConfig=waiter_config,
                )
            except Exception as e:
                result["comment"] = result["comment"] + (str(e),)
                result["result"] = False
            result["comment"] = hub.tool.aws.comment_utils.create_comment(
                resource_type="aws.eks.addon", name=name
            )
        except Exception as e:
            result["comment"] = result["comment"] + (f"{e.__class__.__name__}: {e}",)
            result["result"] = False
            return result

    try:
        if ctx.get("test", False):
            result["new_state"] = plan_state
        elif (not before) or resource_updated:
            after = await hub.exec.boto3.client.eks.describe_addon(
                ctx, clusterName=cluster_name, addonName=resource_id
            )
            result[
                "new_state"
            ] = hub.tool.aws.eks.conversion_utils.convert_raw_addon_to_present(
                raw_resource=after["ret"]["addon"], idem_resource_name=name
            )
        else:
            result["new_state"] = copy.deepcopy(result["old_state"])
    except Exception as e:
        result["comment"] = result["comment"] + (str(e),)
        result["result"] = False
    return result


async def absent(
    hub,
    ctx,
    name: str,
    cluster_name: str,
    resource_id: str = None,
    preserve: bool = False,
    timeout: Dict = None,
) -> Dict[str, Any]:
    r"""
    **Autogenerated function**

    Delete an Amazon EKS add-on. When you remove the add-on, it will also be deleted from the cluster. You can
    always manually start an add-on on the cluster using the Kubernetes API.

    Args:
        name(Text): An Idem name of the EKS addon resource.
        cluster_name(Text): The name of the cluster to delete the add-on from.
        resource_id(Text, optional): AWS EKS addon name. Idem automatically considers this resource being absent
         if this field is not specified.
        preserve(Boolean, optional): Specifying this option preserves the add-on software on your cluster but Amazon
            EKS stops managing any settings for the add-on. If an IAM account is associated with the add-on,
            it is not removed.
        timeout(Dict, optional): Timeout configuration for creating or updating cluster.
            * delete (Dict) -- Timeout configuration for deleting cluster
                * delay -- The amount of time in seconds to wait between attempts.
                * max_attempts -- Customized timeout configuration containing delay and max attempts.

    Request syntax:

        [idem-eks-addon-name]:
              aws.eks.addon.absent:
              - cluster_name: 'string'
              - resource_id: 'string'
              - preserve: boolean

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: sls

            kube-proxy:
              aws.eks.addon.absent:
              - cluster_name: eks-cluster-1
              - resource_id: kube-proxy
              - preserve: True
    """

    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)
    if not resource_id:
        result["comment"] = hub.tool.aws.comment_utils.already_absent_comment(
            resource_type="aws.eks.addon", name=name
        )
        return result
    before = await hub.exec.boto3.client.eks.describe_addon(
        ctx, clusterName=cluster_name, addonName=resource_id
    )

    if not before["ret"]:
        result["comment"] = hub.tool.aws.comment_utils.already_absent_comment(
            resource_type="aws.eks.addon", name=name
        )
    elif ctx.get("test", False):
        result[
            "old_state"
        ] = hub.tool.aws.eks.conversion_utils.convert_raw_addon_to_present(
            raw_resource=before["ret"]["addon"], idem_resource_name=name
        )
        result["comment"] = hub.tool.aws.comment_utils.would_delete_comment(
            resource_type="aws.eks.addon", name=name
        )
        return result
    else:
        result[
            "old_state"
        ] = hub.tool.aws.eks.conversion_utils.convert_raw_addon_to_present(
            raw_resource=before["ret"]["addon"], idem_resource_name=name
        )
        try:
            ret = await hub.exec.boto3.client.eks.delete_addon(
                ctx, clusterName=cluster_name, addonName=resource_id, preserve=preserve
            )
            result["result"] = ret["result"]
            if not result["result"]:
                result["comment"] = ret["comment"]
                result["result"] = False
                return result
            else:
                waiter_config = hub.tool.aws.waiter_utils.create_waiter_config(
                    default_delay=60,
                    default_max_attempts=40,
                    timeout_config=timeout.get("delete") if timeout else None,
                )
                try:
                    await hub.tool.boto3.client.wait(
                        ctx,
                        "eks",
                        "addon_deleted",
                        addonName=resource_id,
                        clusterName=cluster_name,
                        WaiterConfig=waiter_config,
                    )
                except Exception as e:
                    result["comment"] = result["comment"] + (str(e),)
                    result["result"] = False
            result["comment"] = result[
                "comment"
            ] + hub.tool.aws.comment_utils.delete_comment(
                resource_type="aws.eks.addon", name=name
            )
        except hub.tool.boto3.exception.ClientError as e:
            result["comment"] = result["comment"] + (f"{e.__class__.__name__}: {e}",)

    return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    r"""
    **Autogenerated function**

    Describe the resource in a way that can be recreated/managed with the corresponding "present" function


    Lists the available add-ons.


    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: bash

            $ idem describe aws.eks.addon
    """

    result = {}
    cluster_ret = await hub.exec.boto3.client.eks.list_clusters(ctx)
    if not cluster_ret["result"]:
        hub.log.debug(f"Could not describe cluster {cluster_ret['comment']}")
        return {}
    cluster_addons = []
    try:
        for name in cluster_ret["ret"]["clusters"]:
            addon_ret = await hub.exec.boto3.client.eks.list_addons(
                ctx, clusterName=name
            )
            if not addon_ret["result"]:
                hub.log.debug(
                    f"Could not describe cluster addons for cluster {name}, Error:  {addon_ret['comment']}"
                )
                continue
            for addon in addon_ret["ret"]["addons"]:
                cluster_addons.append({"name": name, "addon": addon})
    except Exception as e:
        result["comment"] = str(e)
        result["result"] = False
        return result

    for cluster_addon in cluster_addons:
        cluster_name = cluster_addon["name"]
        current_addon = cluster_addon["addon"]
        resource_id = f"{cluster_name}-{current_addon}"
        describe_ret = await hub.exec.boto3.client.eks.describe_addon(
            ctx, clusterName=cluster_name, addonName=current_addon
        )
        if not describe_ret["result"] or not describe_ret["ret"]:
            hub.log.debug(
                f"Could not describe addon {name}, Error:  {describe_ret['comment']}"
            )
            continue
        add_on = describe_ret["ret"]["addon"]
        resource_translated = (
            hub.tool.aws.eks.conversion_utils.convert_raw_addon_to_present(
                raw_resource=add_on, idem_resource_name=resource_id
            )
        )
        result[resource_id] = {
            "aws.eks.addon.present": [
                {parameter_key: parameter_value}
                for parameter_key, parameter_value in resource_translated.items()
            ]
        }

    return result



from typing import Any
from typing import Dict
from typing import List
from typing import OrderedDict

from dict_tools import differ


async def search_raw(
    hub,
    ctx,
    name,
    resource_id: str = None,
    load_balancer_arn: str = None,
) -> Dict:
    """
    Fetch an ELBv2 Target Group from AWS. The return will be in the same format as what the boto3 api returns.

    **** Note ****: Users can specify one of the following to filter the results: the ARN of the load balancer,
                    the names of one or more target groups, or the ARNs of one or more target groups.

    Order of precedence of input params while performing search: resource_id, load_balancer_arn and finally name.
        If resource_id is not None, regardless whether the 2 other values are None or not, search always is done by
        resource_id only.
    Args:
        name(Text, optional): The name of the Idem state.
        resource_id(Text, optional): AWS ELBv2 Target Group ARN to identify the resource.
        load_balancer_arn(Text, optional): The Amazon Resource Name (ARN) of the load balancer.

    Returns:
        {"result": True|False, "comment": A message List, "ret": Dict}
    """
    result = dict(comment=[], ret=None, result=True)
    if resource_id:
        ret = await hub.exec.boto3.client.elbv2.describe_target_groups(
            ctx,
            TargetGroupArns=[resource_id],
        )
    elif load_balancer_arn:
        ret = await hub.exec.boto3.client.elbv2.describe_target_groups(
            ctx,
            LoadBalancerArn=load_balancer_arn,
        )
    else:
        ret = await hub.exec.boto3.client.elbv2.describe_target_groups(
            ctx, Names=[name]
        )
    result["result"] = ret["result"]
    result["comment"] = list(ret["comment"])
    result["ret"] = ret["ret"]
    return result


async def update(
    hub,
    ctx,
    name: str,
    current_state: Dict[str, Any],
    input_map: Dict[str, Any],
    plan_state: Dict[str, Any],
    resource_id: str,
):
    r"""
    1. Modifies the health checks used when evaluating the health state of the targets in the specified target group.
       If the protocol of the target group is TCP, TLS, UDP, or TCP_UDP, you can't modify the health check protocol,
       interval, timeout, or success codes.
    2. Modifies the specified attributes of the specified target group.
    3. De-registers the specified targets from the specified target group. After the targets are de-registered, they no
       longer receive traffic from the load balancer.
    4. Registers the specified targets with the specified target group. If the target is an EC2 instance, it must be in
       the running state when you register it.
       a. By default, load balancer routes requests to registered targets using the protocol and port for the target group.
          Alternatively, you can override port for a target when you register it.
       b. You can register each EC2 instance or IP address with the same target group multiple times using different ports.
          With a Network Load Balancer, you cannot register instances by instance ID if they have the following instance
          types: C1, CC1, CC2, CG1, CG2, CR1, CS1, G1, G2, HI1, HS1, M1, M2, M3, and T1. You can register instances of
          these types by IP address.

    Args:
        name(Text):The name of the AWS ElasticLoadBalancingv2 target group.
        current_state(Dict[str, Any]): response returned by describe on an AWS ELBv2 Target Group.
        input_map(Dict[st, Any]): a dictionary with newly passed values of params.
        resource_id(Text): AWS ElasticLoadBalancingv2 target group ARN.
        plan_state(Dict[str, Any]): idem --test state for update on AWS ElasticLoadBalancingv2 target group.

    Returns:
        {"result": True|False, "comment": A message List, "ret": None}

    """
    result = dict(comment=[], result=True, ret=[])
    if not ctx.get("test", False):
        if input_map:
            ret = compare_config(current_state=current_state, updated_state=input_map)
            if ret["result"] and ret["ret"]:
                modify_target_group = (
                    await hub.exec.boto3.client.elbv2.modify_target_group(
                        ctx,
                        TargetGroupArn=resource_id,
                        HealthCheckProtocol=ret["ret"].get("HealthCheckProtocol"),
                        HealthCheckPort=ret["ret"].get("HealthCheckPort"),
                        HealthCheckPath=ret["ret"].get("HealthCheckPath"),
                        HealthCheckEnabled=ret["ret"].get("HealthCheckEnabled"),
                        HealthCheckIntervalSeconds=ret["ret"].get(
                            "HealthCheckIntervalSeconds"
                        ),
                        HealthCheckTimeoutSeconds=ret["ret"].get(
                            "HealthCheckTimeoutSeconds"
                        ),
                        HealthyThresholdCount=ret["ret"].get("HealthyThresholdCount"),
                        UnhealthyThresholdCount=ret["ret"].get(
                            "UnhealthyThresholdCount"
                        ),
                        Matcher=ret["ret"].get("Matcher"),
                    )
                )
                if not modify_target_group["result"]:
                    result["comment"] = list(modify_target_group["comment"])
                    result["result"] = False
                    return result
                result["comment"] = list(
                    hub.tool.aws.comment_utils.update_comment(
                        resource_type="aws.elbv2.target_group", name=name
                    )
                )
                result["ret"].append(
                    {
                        "modify_target_group": modify_target_group["ret"][
                            "TargetGroups"
                        ][0]
                    }
                )

        if input_map.get("attributes"):
            modify_attributes = (
                await hub.exec.boto3.client.elbv2.modify_target_group_attributes(
                    ctx,
                    TargetGroupArn=resource_id,
                    Attributes=input_map.get("attributes"),
                )
            )
            if not modify_attributes["result"]:
                result["comment"] += list(modify_attributes["comment"])
                result["result"] = False
                return result
            result["comment"].append("Modified Attributes.")
            result["ret"].append(
                {"modify_attributes": modify_attributes["ret"]["Attributes"]}
            )

        if input_map.get("targets"):
            ret = compare_targets(
                old_targets=current_state.get("targets"),
                new_targets=input_map.get("targets"),
            )
            if ret["ret"]:
                if ret["ret"]["to_deregister"]:
                    deregister_targets = (
                        await hub.exec.boto3.client.elbv2.deregister_targets(
                            ctx,
                            TargetGroupArn=resource_id,
                            Targets=ret["ret"]["to_deregister"],
                        )
                    )
                    if not deregister_targets["result"]:
                        result["comment"] += list(deregister_targets["comment"])
                        result["result"] = False
                        return result
                    result["comment"].append("Deregistered Targets.")
                    result["ret"].append(
                        {"deregister_targets": deregister_targets["ret"]}
                    )

                if ret["ret"]["to_register"]:
                    register_targets = (
                        await hub.exec.boto3.client.elbv2.register_targets(
                            ctx,
                            TargetGroupArn=resource_id,
                            Targets=ret["ret"]["to_register"],
                        )
                    )
                    if not register_targets["result"]:
                        result["comment"] += list(register_targets["comment"])
                        result["result"] = False
                        return result
                    result["comment"].append("Registered Targets.")
                    result["ret"].append({"register_targets": register_targets["ret"]})
    else:
        update_params = OrderedDict(
            {
                "name": name,
                "protocol": input_map.get("protocol"),
                "protocol_version": input_map.get("protocol_version"),
                "port": input_map.get("port"),
                "vpc_id": input_map.get("vpc_id"),
                "health_check_protocol": input_map.get("health_check_protocol"),
                "health_check_port": input_map.get("health_check_port"),
                "health_check_enabled": input_map.get("health_check_enabled"),
                "health_check_path": input_map.get("health_check_path"),
                "health_check_interval_seconds": input_map.get(
                    "health_check_interval_seconds"
                ),
                "health_check_timeout_seconds": input_map.get(
                    "health_check_timeout_seconds"
                ),
                "healthy_threshold_count": input_map.get("healthy_threshold_count"),
                "unhealthy_threshold_count": input_map.get("unhealthy_threshold_count"),
                "matcher": input_map.get("matcher"),
                "target_type": input_map.get("target_type"),
                "tags": input_map.get("tags"),
                "ip_address_type": input_map.get("ip_address_type"),
                "resource_id": resource_id,
                "targets": input_map.get("targets"),
            }
        )
        for key, value in update_params.items():
            if value is not None:
                plan_state[key] = value
        result["ret"] = plan_state
    return result


def compare_config(current_state: Dict[str, Any], updated_state: Dict[str, Any]):
    """
    Compares target group's existing configuration with newly passed input config and returns details of the diff.

    Args:
        current_state (Dict[str, Any]): response returned by describe on an AWS ELBv2 Target Group
        updated_state (Dict[st, Any]): a dictionary with newly passed values of params.

    Returns:
        {"result": True|False, "ret": Dict}

    """
    result = dict(result=False, ret=None)
    target_group_params = {
        "HealthCheckProtocol": "health_check_protocol",
        "HealthCheckPort": "health_check_port",
        "HealthCheckPath": "health_check_path",
        "HealthCheckEnabled": "health_check_enabled",
        "HealthCheckIntervalSeconds": "health_check_interval_seconds",
        "HealthCheckTimeoutSeconds": "health_check_timeout_seconds",
        "HealthyThresholdCount": "healthy_threshold_count",
        "UnhealthyThresholdCount": "unhealthy_threshold_count",
        "Matcher": "matcher",
    }
    current_config_mapping = {}
    for name, value in target_group_params.items():
        if current_state.get(value):
            current_config_mapping[name] = current_state.get(value)

    updated_config_mapping = {}
    for name, value in target_group_params.items():
        if updated_state.get(value):
            updated_config_mapping[name] = updated_state.get(value)

    diff_in_config = differ.deep_diff(current_config_mapping, updated_config_mapping)
    if diff_in_config.get("new"):
        result["result"] = True
        result["ret"] = diff_in_config.get("new")
    return result


def compare_targets(
    old_targets: List[Dict[str, Any]], new_targets: List[Dict[str, Any]]
):
    """
    1. Compares old_targets and new_targets and return the new list of targets that need to be created/ updated.
    2. After targets are de-registered, they no longer receive traffic from the load balancer.
    3. If you specified a port-override when you registered a target, you must specify both the target ID and the port
        when you de-register it.

    Args:
        new_targets(List[Dict[str, Any]]): Newer list of targets to be registered with ELB v2 Target Group.
        old_targets(List[Dict[str, Any]]): Existing list of listeners de-registered with ELB v2 Target Group.

    Returns:
        {"comment": ("A tuple",), "ret": Dict}

    """
    result = dict(ret=None)
    to_deregister = []
    to_register = []
    old_targets_map = {target.get("Id"): target for target in old_targets or []}
    if new_targets is not None:
        for target in new_targets:
            if target.get("Id") in old_targets_map:
                if target == old_targets_map[target.get("Id")]:
                    del old_targets_map[target.get("Id")]
                else:
                    to_register.append(target)
            else:
                to_register.append(target)

        to_deregister = list(old_targets_map.values())
        result["ret"] = {"to_register": to_register, "to_deregister": to_deregister}
    return result
